# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Tests for our CLI applications."""

import contextlib

from click.testing import CliRunner


@contextlib.contextmanager
def stdout_logging():
    # copy logging messages to std out

    import io
    import logging

    buf = io.StringIO()
    ch = logging.StreamHandler(buf)
    ch.setFormatter(logging.Formatter("%(message)s"))
    ch.setLevel(logging.INFO)
    logger = logging.getLogger("mednet")
    logger.addHandler(ch)
    yield buf
    logger.removeHandler(ch)


def _assert_exit_0(result):
    assert (
        result.exit_code == 0
    ), f"Exit code {result.exit_code} != 0 -- Output:\n{result.output}"


def _check_help(entry_point):
    runner = CliRunner()
    result = runner.invoke(entry_point, ["--help"])
    _assert_exit_0(result)
    assert result.output.startswith("Usage:")


def test_info_help():
    from mednet.scripts.info import info

    _check_help(info)


def test_info():
    from mednet.scripts.info import info

    runner = CliRunner()
    result = runner.invoke(info)
    _assert_exit_0(result)
    assert "platform:" in result.output
    assert "accelerators:" in result.output
    assert "version:" in result.output
    assert "databases:" in result.output
    assert "dependencies:" in result.output
    assert "python:" in result.output


def test_config_help():
    from mednet.scripts.config import config

    _check_help(config)


def test_config_list_help():
    from mednet.scripts.config import list_

    _check_help(list_)


def test_config_list():
    from mednet.scripts.config import list_

    runner = CliRunner()
    result = runner.invoke(list_)
    _assert_exit_0(result)
    assert "module: mednet.config.classify.data" in result.output
    assert "module: mednet.config.classify.models" in result.output
    assert "module: mednet.config.segment.data" in result.output
    assert "module: mednet.config.segment.models" in result.output


def test_config_list_v():
    from mednet.scripts.config import list_

    result = CliRunner().invoke(list_, ["--verbose"])
    _assert_exit_0(result)
    assert "module: mednet.config.classify.data" in result.output
    assert "module: mednet.config.classify.models" in result.output
    assert "module: mednet.config.segment.data" in result.output
    assert "module: mednet.config.segment.models" in result.output


def test_config_describe_help():
    from mednet.scripts.config import describe

    _check_help(describe)


def test_database_help():
    from mednet.scripts.database import database

    _check_help(database)


def test_database_list_help():
    from mednet.scripts.database import list_

    _check_help(list_)


def test_database_list():
    from mednet.scripts.database import list_

    runner = CliRunner()
    result = runner.invoke(list_)
    _assert_exit_0(result)
    assert result.output.startswith("  - ")


def test_database_check_help():
    from mednet.scripts.database import check

    _check_help(check)


def test_database_preprocess_help():
    from mednet.scripts.preprocess import preprocess

    _check_help(preprocess)


def test_train_help():
    from mednet.scripts.train import train

    _check_help(train)


def test_predict_help():
    from mednet.scripts.predict import predict

    _check_help(predict)


def test_experiment_help():
    from mednet.scripts.experiment import experiment

    _check_help(experiment)


def test_upload_help():
    from mednet.scripts.upload import upload

    _check_help(upload)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import json
import pathlib
import typing

import click
import tqdm
from clapper.click import ResourceOption, verbosity_option

from ...engine.segment.evaluator import SUPPORTED_METRIC_TYPE
from ..click import ConfigCommand
from ..logging import setup_cli_logger

logger = setup_cli_logger()


@click.command(
    entry_point_group="mednet.config",
    cls=ConfigCommand,
    epilog="""Examples:

\b
  1. Produces visualisations for the given predictions:

     .. code:: sh

        $ mednet segment view -vv --predictions=path/to/predictions.json --output-folder=path/to/results

  2. Produces visualisations for the given (dumped) annotations:

     .. code:: sh

        $ mednet segment view -vv --predictions=path/to/annotations.json --output-folder=path/to/results
""",
)
@click.option(
    "--predictions",
    "-p",
    help="""Path to the JSON file describing available predictions. The actual
    predictions are supposed to lie on the same folder.""",
    required=True,
    type=click.Path(
        file_okay=True,
        dir_okay=False,
        writable=False,
        path_type=pathlib.Path,
    ),
    show_default=True,
    cls=ResourceOption,
)
@click.option(
    "--output-folder",
    "-o",
    help="Directory in which to store results (created if does not exist)",
    required=True,
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        writable=True,
        path_type=pathlib.Path,
    ),
    default="results",
    cls=ResourceOption,
)
@click.option(
    "--threshold",
    "-t",
    help="""This number is used to define positives and negatives from
    probability maps, and used to report metrics based on a threshold chosen *a
    priori*. It can be set to a floating-point value, or to the name of dataset
    split in ``--predictions``.
    """,
    default="0.5",
    show_default=True,
    required=False,
    cls=ResourceOption,
)
@click.option(
    "--metric",
    "-m",
    help="""If threshold is set to the name of a split in ``--predictions``,
    then this parameter defines the metric function to be used to evaluate the
    threshold at which the metric reaches its maximum value. All other splits
    are evaluated with respect to this threshold.""",
    default="f1",
    type=click.Choice(typing.get_args(SUPPORTED_METRIC_TYPE), case_sensitive=True),
    show_default=True,
    required=True,
    cls=ResourceOption,
)
@click.option(
    "--steps",
    "-s",
    help="""Number of steps for evaluating metrics on various splits. This
    value is used only when evaluating thresholds on a datamodule split.""",
    default=100,
    type=click.IntRange(10),
    show_default=True,
    required=True,
    cls=ResourceOption,
)
@click.option(
    "--show-errors/--no-show-errors",
    "-e/-E",
    help="""If set, then shows a colorized version of the segmentation map in
    which false-positives are marked in red, and false-negatives in green.
    True positives are always marked in white.""",
    default=False,
    show_default=True,
    required=True,
    cls=ResourceOption,
)
@click.option(
    "--alpha",
    "-a",
    help="""Defines the transparency weighting between the original image and
    the predicted segmentation maps. A value of 1.0 makes the program output
    only segmentation maps.  A value of 0.0 makes the program output only the
    processed image.""",
    default=0.6,
    type=click.FloatRange(0.0, 1.0),
    show_default=True,
    required=True,
    cls=ResourceOption,
)
@verbosity_option(logger=logger, expose_value=False)
def view(
    predictions: pathlib.Path,
    output_folder: pathlib.Path,
    threshold: str | float,
    metric: str,
    steps: int,
    show_errors: bool,
    alpha: float,
    **_,  # ignored
):  # numpydoc ignore=PR01
    """Display predictions generated by a model."""

    import numpy

    from ...engine.segment.evaluator import (
        compute_metric,
        load_count,
        name2metric,
        validate_threshold,
    )
    from ...engine.segment.viewer import view
    from ..utils import execution_metadata, save_json_with_backup

    view_filename = "view.json"
    view_file = output_folder / view_filename

    with predictions.open("r") as f:
        predict_data = json.load(f)

    # register metadata
    json_data: dict[str, typing.Any] = execution_metadata()
    json_data.update(
        dict(
            predictions=str(predictions),
            output_folder=str(output_folder),
            threshold=threshold,
            metric=metric,
            steps=steps,
        ),
    )
    json_data = {k.replace("_", "-"): v for k, v in json_data.items()}
    save_json_with_backup(view_file.with_suffix(".meta.json"), json_data)

    threshold = validate_threshold(threshold, predict_data)
    threshold_list = numpy.arange(
        0.0, (1.0 + 1 / steps), 1 / steps, dtype=numpy.float64
    )

    if isinstance(threshold, str):
        # Compute threshold on specified split, if required
        logger.info(f"Evaluating threshold on `{threshold}` split using " f"`{metric}`")
        counts = load_count(predictions.parent, predict_data[threshold], threshold_list)
        metric_list = compute_metric(
            counts, name2metric(typing.cast(SUPPORTED_METRIC_TYPE, metric))
        )
        threshold_index = metric_list.argmax()
        logger.info(f"Set --threshold={threshold_list[threshold_index]:.4f}")

    else:
        # must figure out the closest threshold from the list we are using
        threshold_index = (numpy.abs(threshold_list - threshold)).argmin()
        logger.info(f"Set --threshold={threshold_list[threshold_index]:.4f}")

    # create visualisations
    for split_name, sample_list in predict_data.items():
        logger.info(
            f"Creating {len(sample_list)} visualisations for split `{split_name}`"
        )
        for sample in tqdm.tqdm(sample_list):
            image = view(
                predictions.parent,
                sample[1],
                threshold=threshold_list[threshold_index],
                show_errors=show_errors,
                tp_color=(255, 255, 255),
                fp_color=(255, 0, 0),
                fn_color=(0, 255, 0),
                alpha=alpha,
            )
            dest = (output_folder / sample[1]).with_suffix(".png")
            dest.parent.mkdir(parents=True, exist_ok=True)
            tqdm.tqdm.write(f"{sample[1]} -> {dest}")
            image.save(dest)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Defines functionality for the evaluation of object detection predictions."""

import collections
import logging
import typing

import matplotlib.axes
import matplotlib.figure
import numpy
import numpy.typing
import tabulate
import torch
import torchvision.ops
from matplotlib import pyplot as plt

from ...models.detect.typing import Prediction

logger = logging.getLogger(__name__)


def _compute_iou_from_predictions(
    predictions: typing.Sequence[Prediction],
) -> list[list[tuple[int, float, int, float]]]:
    """Calculate the IOU for each **detected** bounding-box in predictions.

    This function will calculate the IOU (intersection over union) metric for each
    detected bounding-box (as in output by a model) in the prediction dataset.  It will
    then return a list of tuples, each matching a prediction, indicating the matched
    target, the IOU, the class and the model score.

    Parameters
    ----------
    predictions
        A list of predictions to consider for measurement.

    Returns
    -------
        A list containing lists of tuples, matching the order of **detected**
        bounding-boxes (as in output by a model). Each tuple contains the index of the
        matching target bounding box, the IOU between the target and said detected
        bounding-box, the class of the said target/detected bounding box, and finally
        the model score.

        In case there is no match for a particular detected bounding-box, the output
        table matching this would contain ``(-1, 0.0, <class>, 0.0)``.  This model
        output can be accounted as a "misdetection".
    """

    retval: list[list[tuple[int, float, int, float]]] = []

    for sample in predictions:
        name, targets, detections = sample

        # calculates IOU of all targets against all bounding boxes
        if detections:
            iou = torchvision.ops.box_iou(
                torch.tensor([k[0] for k in targets]),
                torch.tensor([k[0] for k in detections]),
            ).numpy()
        else:
            logger.warning(f"No detections for sample `{name}` were found.")
            retval.append(list())
            continue

        # we are only interested in the positions in which targets and detections have
        # matching labels - everything else can be set to zero on the IOU matrix
        iou *= numpy.equal.outer(
            [k[1] for k in targets], [k[1] for k in detections]
        ).astype(int)

        # the order of attributions need to go from highest to lowest score
        attribution_order = numpy.flip(numpy.argsort([k[2] for k in detections]))

        attributions: list[tuple[int, float, int, float]] = []
        for detection_index in attribution_order:
            max_iou_arg = iou[:, detection_index].argmax()
            if iou[max_iou_arg, detection_index] > 0.0:  # match
                attributions.append(
                    (
                        max_iou_arg,
                        iou[max_iou_arg, detection_index],
                        detections[detection_index][1],
                        detections[detection_index][2],
                    )
                )
                iou[max_iou_arg] = 0.0  # this (ground-truth) target has been attributed
            else:  # no match
                attributions.append((-1, 0.0, detections[detection_index][1], 0.0))

        retval.append(attributions)

    return retval


def run(
    predictions: typing.Sequence[Prediction],
    binning: str | int,
    iou_threshold: float | None = None,
) -> dict[str, typing.Any]:
    """Run inference and calculates measures for multilabel object detection.

    Parameters
    ----------
    predictions
        A list of predictions to consider for measurement.
    binning
        The binning algorithm to use for computing the bin widths and
        distribution for histograms.  Choose from algorithms supported by
        :py:func:`numpy.histogram`.
    iou_threshold
        IOU threshold by which we consider successful object detection. If set to
        ``None``, then apply no thresholding.

    Returns
    -------
        A dictionary containing the performance summary on the specified threshold,
        general performance curves (under the key ``curves``), and score histograms
        (under the key ``score-histograms``).
    """

    detailed_iou = _compute_iou_from_predictions(predictions)
    if iou_threshold is not None:
        filtered_iou = [[k for k in j if k[1] >= iou_threshold] for j in detailed_iou]
    else:
        filtered_iou = detailed_iou

    iou_histogram = dict(
        zip(
            ("hist", "bin_edges"),
            numpy.histogram(
                [k[1] for j in filtered_iou for k in j if k], bins=binning, range=(0, 1)
            ),
        )
    )

    # the mean-iou only accounts the IoU for matches - non-matches are ignored on this
    # metric (test: k[0] >= 0)
    mean_iou = numpy.nan_to_num(
        numpy.mean([k[1] for j in filtered_iou for k in j if k if k[0] >= 0])
    )
    classes = sorted(list(set([k[1] for j in predictions for k in j[1]])))

    # the mean-iou only accounts the IoU for matches - non-matches are ignored on this
    # metric (test: k[0] >= 0)
    mean_iou_per_class = {
        cl: numpy.nan_to_num(
            numpy.mean(
                [
                    k[1]
                    for j in filtered_iou
                    for k in j
                    if k and k[0] >= 0 and k[2] == cl
                ]
            )
        )
        for cl in classes
    }
    num_targets_per_class = collections.Counter(
        [k[1] for j in predictions for k in j[1]]
    )
    num_detections_per_class = collections.Counter(
        [k[2] for j in filtered_iou for k in j if k]
    )
    iou_per_class = [(k[2], k[1]) for j in filtered_iou for k in j if k]
    iou_per_class = {cl: [k[1] for k in iou_per_class if k[0] == cl] for cl in classes}  # type: ignore

    per_class_histogram = {
        cl: dict(
            zip(
                ("hist", "bin_edges"),
                numpy.histogram(iou_per_class[cl], bins=binning, range=(0, 1)),
            )
        )
        for cl in classes
    }

    return {
        "num-samples": len(predictions),
        "num-targets": sum([len(k[1]) for k in predictions]),
        "num-detections": len([k[2] for j in filtered_iou for k in j if k]),
        "mean-iou": mean_iou,
        "iou-histogram": iou_histogram,
        "per-class": {
            cl: {
                "mean-iou": mean_iou_per_class[cl],
                "num-targets": num_targets_per_class[cl],
                "num-detections": num_detections_per_class[cl],
            }
            for cl in classes
        },
        "per-class-iou-histogram": per_class_histogram,
    }


def make_table(
    data: typing.Mapping[str, typing.Mapping[str, typing.Any]],
    fmt: str,
) -> str:
    """Tabulate summaries from multiple splits.

    This function can properly tabulate the various summaries produced for all
    the splits in a prediction database.

    Parameters
    ----------
    data
        An iterable over all summary data collected.
    fmt
        One of the formats supported by `python-tabulate
        <https://pypi.org/project/tabulate/>`_.

    Returns
    -------
    str
        A string containing the tabulated information.
    """

    def _exclusion_condition(v: str) -> bool:
        return v not in ("iou-histogram", "per-class", "per-class-iou-histogram")

    # dump evaluation results in RST format to screen and file
    table_data = {}
    for k, v in data.items():
        table_data[k] = {kk: vv for kk, vv in v.items() if _exclusion_condition(kk)}

    example = next(iter(table_data.values()))
    headers = list(example.keys())
    table = [[k[h] for h in headers] for k in table_data.values()]

    # add subset names
    headers = ["subset"] + headers
    table = [[name] + k for name, k in zip(table_data.keys(), table)]

    return tabulate.tabulate(table, headers, tablefmt=fmt, floatfmt=".3f")


def _iou_histogram_plot(
    histograms: dict[str, numpy.typing.NDArray]
    | dict[int, dict[str, numpy.typing.NDArray]],
    title: str,
    threshold: float | None,
) -> matplotlib.figure.Figure:
    """Plot the normalized score distributions for all systems.

    Parameters
    ----------
    histograms
        A dictionary containing all histograms that should be inserted into the
        plot.  Each histogram should itself be setup as another dictionary
        containing the keys ``hist`` and ``bin_edges`` as returned by
        :py:func:`numpy.histogram`.
    title
        Title of the plot.
    threshold
        Shows where the threshold is in the figure.  If set to ``None``, then
        does not show the threshold line.

    Returns
    -------
    matplotlib.figure.Figure
        A single (matplotlib) plot containing the score distribution, ready to
        be saved to disk or displayed.
    """

    from matplotlib.ticker import MaxNLocator

    fig, ax = plt.subplots(1, 1)
    assert isinstance(fig, matplotlib.figure.Figure)
    ax = typing.cast(matplotlib.axes.Axes, ax)  # gets editor to behave

    # Here, we configure the "style" of our plot
    ax.set_xlim((0, 1))
    ax.set_title(title)
    ax.set_xlabel("Score")
    ax.set_ylabel("Count")

    # Only show ticks on the left and bottom spines
    ax.spines.right.set_visible(False)
    ax.spines.top.set_visible(False)
    ax.get_xaxis().tick_bottom()
    ax.get_yaxis().tick_left()
    ax.get_yaxis().set_major_locator(MaxNLocator(integer=True))

    # Setup the grid
    ax.grid(linestyle="--", linewidth=1, color="gray", alpha=0.2)
    ax.get_xaxis().grid(False)

    max_hist = 0
    has_labels = False

    if isinstance(next(iter(histograms.keys())), int):
        histograms = typing.cast(dict[int, dict[str, numpy.typing.NDArray]], histograms)
        # per-class diagram, requires labels
        for cl in histograms.keys():
            hist = histograms[cl]["hist"]
            bin_edges = histograms[cl]["bin_edges"]
            width = 0.7 * (bin_edges[1] - bin_edges[0])
            center = (bin_edges[:-1] + bin_edges[1:]) / 2
            ax.bar(center, hist, align="center", width=width, label=str(cl), alpha=0.7)
            max_hist = max(max_hist, hist.max())
            has_labels |= True
    else:
        # single histogram, no need for labels
        histograms = typing.cast(dict[str, numpy.typing.NDArray], histograms)
        hist = histograms["hist"]
        bin_edges = histograms["bin_edges"]
        width = 0.7 * (bin_edges[1] - bin_edges[0])
        center = (bin_edges[:-1] + bin_edges[1:]) / 2
        ax.bar(center, hist, align="center", width=width, alpha=0.7)
        max_hist = max(max_hist, hist.max())
        has_labels |= False

    # Detach axes from the plot
    ax.spines["left"].set_position(("data", -0.015))
    ax.spines["bottom"].set_position(("data", -0.015 * max_hist))

    if threshold is not None:
        # Adds threshold line (dotted red)
        ax.axvline(
            threshold,  # type: ignore
            color="red",
            lw=2,
            alpha=0.75,
            ls="dotted",
            label="IOU threshold",
        )
        has_labels |= True

    # Adds a nice legend
    if has_labels:
        ax.legend(
            fancybox=True,
            framealpha=0.7,
        )

    # Makes sure the figure occupies most of the possible space
    fig.tight_layout()

    return fig


def make_plots(
    results: dict[str, dict[str, typing.Any]], iou_threshold: float | None = None
) -> list:
    """Create plots for all curves and score distributions in ``results``.

    Parameters
    ----------
    results
        Evaluation data as returned by :py:func:`run`.
    iou_threshold
        IOU threshold by which we consider successful object detection.  If set, it is
        shown on plots.

    Returns
    -------
        A list of figures to record to file
    """

    retval = []

    # score plots
    for split_name, data in results.items():
        retval.append(
            _iou_histogram_plot(
                data["iou-histogram"],
                f"IOU distribution (split: {split_name})",
                threshold=iou_threshold,
            )
        )

        if len(data["per-class-iou-histogram"]) > 1:
            retval.append(
                _iou_histogram_plot(
                    data["per-class-iou-histogram"],
                    f"IOU distribution per class (split: {split_name})",
                    threshold=iou_threshold,
                )
            )

    return retval



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Drishti-GS1 for Optic Disc and Cup Segmentation.

Drishti-GS is a dataset meant for validation of segmenting OD, cup and
detecting notching.  The images in the Drishti-GS dataset have been collected
and annotated by Aravind Eye hospital, Madurai, India. This dataset is of a
single population as all subjects whose eye images are part of this dataset are
Indians.

The dataset is divided into two: a training set and a testing set of images.
Training images (50) are provided with groundtruths for OD and Cup segmentation
and notching information.

* Reference (including train/test split): :cite:p:`sivaswamy_drishti-gs_2014`
* Original resolution (height x width): varying (min: 1749x2045 pixels,
  max: 1845x2468 pixels)
* Protocols ``optic-disc`` and ``optic-cup``:

  * Training: 50
  * Test: 51

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import importlib.resources
import importlib.resources.abc
import os
import pathlib
import typing

import PIL.Image
import torch
from torchvision import tv_tensors
from torchvision.transforms.v2.functional import to_dtype, to_image

from ...models.transforms import crop_image_to_mask
from ...utils.rc import load_rc
from ..datamodule import CachingDataModule
from ..split import JSONDatabaseSplit
from ..typing import RawDataLoader as BaseDataLoader
from ..typing import Sample

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name to refer to this database."""

CONFIGURATION_KEY_DATADIR = "datadir." + DATABASE_SLUG
"""Key to search for in the configuration file for the root directory of this
database."""


class RawDataLoader(BaseDataLoader):
    """A specialized raw-data-loader for the drishtigs1 dataset.

    Parameters
    ----------
    target_all
        Indicate whether to use the "all" or "any" target.
    """

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self, target_all: bool):
        self.datadir = pathlib.Path(
            load_rc().get(CONFIGURATION_KEY_DATADIR, os.path.realpath(os.curdir))
        )

        self.target_all = target_all

    def sample(self, sample: typing.Any) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing path suffixes to the sample image, target, and mask
            to be loaded, within the dataset root folder.

        Returns
        -------
            The sample representation.
        """

        image = PIL.Image.open(self.datadir / sample[0]).convert(mode="RGB")
        image = to_dtype(to_image(image), torch.float32, scale=True)

        target = self.target(sample)

        mask_path = (
            importlib.resources.files(__package__) / "masks" / DATABASE_SLUG / sample[2]
        )
        with importlib.resources.as_file(mask_path) as path:
            mask = PIL.Image.open(path).convert(mode="1", dither=None)
            mask = to_dtype(to_image(mask), torch.float32, scale=True)

        image = tv_tensors.Image(crop_image_to_mask(image, mask))
        target = tv_tensors.Mask(crop_image_to_mask(target, mask))
        mask = tv_tensors.Mask(mask)

        return dict(image=image, target=target, mask=mask, name=sample[0])

    def target(self, sample: typing.Any) -> torch.Tensor:
        """Load only sample target from its raw representation.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The label corresponding to the specified sample, encapsulated as a
            torch float tensor.
        """

        if self.target_all:
            target = (
                PIL.Image.open(self.datadir / sample[1])
                .convert(mode="RGB", dither=None)
                .convert("L")
                .point(lambda p: p > 254, mode="1")
            )

        else:
            target = (
                PIL.Image.open(self.datadir / sample[1])
                .convert(mode="RGB", dither=None)
                .convert("L")
                .point(lambda p: p > 0, mode="1")
            )

        return to_dtype(to_image(target), torch.float32, scale=True)


class DataModule(CachingDataModule):
    """Drishti-GS1 for Optic Disc and Cup Segmentation.

    Parameters
    ----------
    split_path
        Path or traversable (resource) with the JSON split description to load.
    target_all
        Indicate whether to use the "all" or "any" target.
    """

    def __init__(
        self,
        split_path: pathlib.Path | importlib.resources.abc.Traversable,
        target_all: bool,
    ):
        super().__init__(
            database_split=JSONDatabaseSplit(split_path),
            raw_data_loader=RawDataLoader(target_all),
            database_name=DATABASE_SLUG,
            split_name=split_path.name.rsplit(".", 2)[0],
            task="segmentation",
            num_classes=1,
        )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Tests for image utilities."""

import numpy
import PIL.Image

from mednet.data.image_utils import remove_black_borders


def test_remove_black_borders(datadir):
    # Get a raw sample with black border
    data_file = str(datadir / "raw_with_black_border.png")
    raw_with_black_border = PIL.Image.open(data_file)

    # Remove the black border
    raw_rbb_removed, _ = remove_black_borders(raw_with_black_border)

    # Get the same sample without black border
    data_file_2 = str(datadir / "raw_without_black_border.png")
    raw_without_black_border = PIL.Image.open(data_file_2)

    # Compare both
    raw_rbb_removed = numpy.asarray(raw_rbb_removed)
    raw_without_black_border = numpy.asarray(raw_without_black_border)

    numpy.testing.assert_array_equal(raw_without_black_border, raw_rbb_removed)


def test_load_pil_16bit(datadir):
    # If the ratio is higher 0.5, image is probably clipped

    image = PIL.Image.open(datadir / "16bits.png")
    array = numpy.array(image).astype(numpy.float32) / 65535

    count_pixels = numpy.count_nonzero(array)
    count_max_value = numpy.count_nonzero(array == array.max())

    assert count_max_value / count_pixels < 0.5



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
import numpy as np
from torchvision import tv_tensors

from mednet.engine.classify.saliency.interpretability import (
    _compute_avg_saliency_focus,
    _compute_binary_mask,
    _compute_proportional_energy,
    _process_sample,
)


def test_compute_avg_saliency_focus():
    grayscale_cams = np.ones((200, 200))
    grayscale_cams2 = np.full((512, 512), 0.5)
    grayscale_cams3 = np.zeros((256, 256))
    grayscale_cams3[50:75, 50:100] = 1

    bbox_data = [50, 50, 100, 100]

    gt_boxes = tv_tensors.BoundingBoxes(
        data=bbox_data, format="XYXY", canvas_size=grayscale_cams.shape
    )
    binary_mask = _compute_binary_mask(gt_boxes, grayscale_cams)
    gt_boxes2 = tv_tensors.BoundingBoxes(
        data=bbox_data, format="XYXY", canvas_size=grayscale_cams.shape
    )
    binary_mask2 = _compute_binary_mask(gt_boxes2, grayscale_cams2)
    gt_boxes3 = tv_tensors.BoundingBoxes(
        data=bbox_data, format="XYXY", canvas_size=grayscale_cams.shape
    )
    binary_mask3 = _compute_binary_mask(gt_boxes3, grayscale_cams3)

    avg_saliency_focus = _compute_avg_saliency_focus(
        grayscale_cams,
        binary_mask,
    )
    avg_saliency_focus2 = _compute_avg_saliency_focus(
        grayscale_cams2,
        binary_mask2,
    )
    avg_saliency_focus3 = _compute_avg_saliency_focus(
        grayscale_cams3,
        binary_mask3,
    )

    assert avg_saliency_focus == 1
    assert avg_saliency_focus2 == 0.5
    assert avg_saliency_focus3 == 0.5


def test_compute_avg_saliency_focus_no_activations():
    grayscale_cams = np.zeros((200, 200))

    bbox_data = [50, 50, 100, 100]

    gt_boxes = tv_tensors.BoundingBoxes(
        data=bbox_data, format="XYXY", canvas_size=grayscale_cams.shape
    )

    binary_mask = _compute_binary_mask(gt_boxes, grayscale_cams)
    avg_saliency_focus = _compute_avg_saliency_focus(
        grayscale_cams,
        binary_mask,
    )

    assert avg_saliency_focus == 0


def test_compute_avg_saliency_focus_zero_gt_area():
    grayscale_cams = np.ones((200, 200))

    bbox_data = [50, 50, 50, 50]

    gt_boxes = tv_tensors.BoundingBoxes(
        data=bbox_data, format="XYXY", canvas_size=grayscale_cams.shape
    )

    binary_mask = _compute_binary_mask(gt_boxes, grayscale_cams)
    avg_saliency_focus = _compute_avg_saliency_focus(
        grayscale_cams,
        binary_mask,
    )

    assert avg_saliency_focus == 0


def test_compute_proportional_energy():
    grayscale_cams = np.ones((200, 200))
    grayscale_cams2 = np.full((512, 512), 0.5)
    grayscale_cams3 = np.zeros((512, 512))
    grayscale_cams3[100:200, 100:200] = 1

    bbox_data = [50, 50, 150, 150]

    gt_boxes = tv_tensors.BoundingBoxes(
        data=bbox_data, format="XYXY", canvas_size=grayscale_cams.shape
    )
    binary_mask = _compute_binary_mask(gt_boxes, grayscale_cams)
    gt_boxes2 = tv_tensors.BoundingBoxes(
        data=bbox_data, format="XYXY", canvas_size=grayscale_cams2.shape
    )
    binary_mask2 = _compute_binary_mask(gt_boxes2, grayscale_cams2)
    gt_boxes3 = tv_tensors.BoundingBoxes(
        data=bbox_data, format="XYXY", canvas_size=grayscale_cams3.shape
    )
    binary_mask3 = _compute_binary_mask(gt_boxes3, grayscale_cams3)

    proportional_energy = _compute_proportional_energy(
        grayscale_cams,
        binary_mask,
    )
    proportional_energy2 = _compute_proportional_energy(
        grayscale_cams2,
        binary_mask2,
    )
    proportional_energy3 = _compute_proportional_energy(
        grayscale_cams3,
        binary_mask3,
    )

    assert proportional_energy == 0.25
    assert proportional_energy2 == 0.03814697265625
    assert proportional_energy3 == 0.25


def test_compute_proportional_energy_no_activations():
    grayscale_cams = np.zeros((200, 200))

    bbox_data = [50, 50, 150, 150]

    gt_boxes = tv_tensors.BoundingBoxes(
        data=bbox_data, format="XYXY", canvas_size=grayscale_cams.shape
    )

    binary_mask = _compute_binary_mask(gt_boxes, grayscale_cams)
    proportional_energy = _compute_proportional_energy(
        grayscale_cams,
        binary_mask,
    )

    assert proportional_energy == 0


def test_compute_proportional_energy_no_gt_box():
    grayscale_cams = np.ones((200, 200))

    bbox_data = [0, 0, 0, 0]

    gt_boxes = tv_tensors.BoundingBoxes(
        data=bbox_data, format="XYXY", canvas_size=grayscale_cams.shape
    )

    binary_mask = _compute_binary_mask(gt_boxes, grayscale_cams)
    proportional_energy = _compute_proportional_energy(
        grayscale_cams,
        binary_mask,
    )

    assert proportional_energy == 0


def test_process_sample():
    grayscale_cams = np.ones((200, 200))

    bbox_data = [50, 50, 50, 50]

    gt_boxes = tv_tensors.BoundingBoxes(
        data=bbox_data, format="XYXY", canvas_size=grayscale_cams.shape
    )

    proportional_energy, avg_saliency_focus = _process_sample(
        gt_boxes,
        grayscale_cams,
    )

    assert proportional_energy == 0
    assert avg_saliency_focus == 0



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib
import sys

import click
from clapper.click import ResourceOption, verbosity_option

from .click import ConfigCommand
from .logging import setup_cli_logger

logger = setup_cli_logger()


@click.command(
    entry_point_group="mednet.config",
    cls=ConfigCommand,
    epilog="""Examples:

1. Pre-process the Montgomery (classification) dataset using model transforms
   from the pasa model.

   .. code:: sh

      mednet database preprocess -vv montgomery pasa

2. Pre-process the CXR8 (segmentation) dataset using model transforms
   from the lwnet model.

   .. code:: sh

      mednet database preprocess -vv cxr8 lwnet

""",
)
@click.option(
    "--output-folder",
    "-o",
    help="Directory in which to store results (created if does not exist)",
    required=True,
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        writable=True,
        path_type=pathlib.Path,
    ),
    default="results",
    show_default=True,
    cls=ResourceOption,
)
@click.option(
    "--model",
    "-m",
    help="A lightning module instance implementing the network to be trained",
    required=True,
    cls=ResourceOption,
)
@click.option(
    "--datamodule",
    "-d",
    help="A lightning DataModule containing the training and validation sets.",
    required=True,
    cls=ResourceOption,
)
@click.option(
    "--limit",
    "-l",
    help="Limit preprocessing to the first N samples in each split in the "
    "configuration, making the job sensibly faster. Set it to "
    "zero (default) to preprocess everything.",
    required=True,
    type=click.IntRange(0),
    default=0,
    show_default=True,
)
@click.option(
    "--grayscale",
    "-g",
    help="For images to be grayscale, if set.",
    required=False,
    is_flag=True,
    show_default=True,
)
@verbosity_option(logger=logger, expose_value=False)
def preprocess(
    output_folder,
    model,
    datamodule,
    limit,
    grayscale,
    **_,
) -> None:  # numpydoc ignore=PR01
    """Run pre-processing on databases using a set of model transforms.

    This command can be used to generate more compact versions of the databases
    that can be used to speed-up data loading. This is done by running both
    data loading and model transforms so that stored samples are small enough,
    and require no transformations when loading.
    """

    from torchvision.transforms.v2.functional import rgb_to_grayscale, to_pil_image

    # report model/transforms options - set data augmentations
    logger.info(f"Network model: {type(model).__module__}.{type(model).__name__}")

    datamodule.model_transforms = model.model_transforms
    datamodule.setup(stage="predict")
    loaders = datamodule.predict_dataloader()

    for split_name, loader in loaders.items():
        if limit == 0:
            click.secho(
                f"Preprocessing all {len(loader)} samples of split "
                f"`{split_name}` at database `{datamodule.database_name}`...",
                fg="yellow",
            )
            loader_limit = sys.maxsize
        else:
            click.secho(
                f"Preprocessing first {limit} samples of dataset "
                f"`{split_name}` at database `{datamodule.database_name}`...",
                fg="yellow",
            )
            loader_limit = limit

        # the for loop will trigger raw data loading (ie. user code), protect it
        for sample_order, batch in enumerate(loader):
            if loader_limit == 0:
                break
            logger.info(
                f"{batch['name'][0]}: "
                f"{[s for s in batch['image'][0].shape]}@{batch['image'][0].dtype}",
            )

            sample_info = datamodule.splits[split_name][0][0][sample_order]

            # basic sanity check
            assert sample_info[0] == batch["name"][0]

            # we are always interested on the images
            output_image = output_folder / sample_info[0]
            output_image.parent.mkdir(parents=True, exist_ok=True)
            tensor_image = batch["image"][0]

            match tensor_image.shape[0]:
                case 1:
                    image = to_pil_image((255 * tensor_image).byte(), "L")
                case 3:
                    if grayscale:
                        image = to_pil_image(
                            (255 * rgb_to_grayscale(tensor_image)).byte(), "L"
                        )
                    else:
                        image = to_pil_image((255 * tensor_image).byte(), "RGB")
                case _:
                    raise NotImplementedError(
                        f"Cannot preprocess images with {tensor_image.shape[1]} planes"
                    )

            image.save(output_image)

            if datamodule.task == "segmentation":
                # we are also interested on the targets
                output_target = output_folder / sample_info[1]
                output_target.parent.mkdir(parents=True, exist_ok=True)
                tensor_target = batch["target"][0]
                target = to_pil_image((255 * tensor_target).byte(), "L")
                target.save(output_target)

                if len(sample_info) >= 3:
                    # we are also interested on the masks
                    output_mask = output_folder / sample_info[2]
                    output_mask.parent.mkdir(parents=True, exist_ok=True)
                    tensor_mask = batch["mask"][0]
                    mask = to_pil_image((255 * tensor_mask).byte(), "L")
                    mask.save(output_mask)

            loader_limit -= 1



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib

import click
from clapper.click import ResourceOption, verbosity_option

from ..click import ConfigCommand
from ..logging import setup_cli_logger

# avoids X11/graphical desktop requirement when creating plots
__import__("matplotlib").use("agg")

logger = setup_cli_logger()


@click.command(
    entry_point_group="mednet.config",
    cls=ConfigCommand,
    epilog="""Examples:

1. Run evaluation on an existing prediction output:

   .. code:: sh

      mednet classify evaluate -vv --predictions=path/to/predictions.json

2. Run evaluation on an existing prediction output, tune threshold a priori on the `validation` set:

   .. code:: sh

      mednet classify evaluate -vv --predictions=path/to/predictions.json --threshold=validation
""",
)
@click.option(
    "--predictions",
    "-p",
    help="Filename in which predictions are currently stored",
    required=True,
    type=click.Path(
        file_okay=True,
        dir_okay=False,
        writable=True,
        path_type=pathlib.Path,
    ),
    cls=ResourceOption,
)
@click.option(
    "--output-folder",
    "-o",
    help="Directory in which to store results (created if does not exist)",
    required=True,
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        writable=True,
        path_type=pathlib.Path,
    ),
    default="results",
    show_default=True,
    cls=ResourceOption,
)
@click.option(
    "--threshold",
    "-t",
    help="""This value is used to define positives and negatives from
    probability outputs in predictions, and report performance measures on
    **binary** classification tasks. It should either come from the training
    set or a separate validation set to avoid biasing the analysis.
    Optionally, if you provide a multi-split set of predictions as input, this
    may also be the name of an existing split (e.g. ``validation``) from which
    the threshold will be estimated (by calculating the threshold leading to
    the highest F1-score on that set) and then applied to the subsequent
    sets.  This value is not used for multi-class classification tasks.""",
    default=0.5,
    show_default=True,
    required=True,
    type=click.STRING,
    cls=ResourceOption,
)
@click.option(
    "--binning",
    "-b",
    help="""The binning algorithm to use for computing the bin widths and
    distribution for histograms.  Choose from algorithms supported by
    :py:func:`numpy.histogram`, or a simple integer indicating the number of
    bins to have in the interval ``[0, 1]``.""",
    default="50",
    show_default=True,
    required=True,
    type=click.STRING,
    cls=ResourceOption,
)
@click.option(
    "--plot/--no-plot",
    "-P",
    help="""If set, then also produces figures containing the plots of
    performance curves and score histograms.""",
    required=True,
    show_default=True,
    default=True,
    cls=ResourceOption,
)
@click.option(
    "--credible-regions/--no-credible-regions",
    "-I",
    help="""If set, then evaluates also credible intervals. Notice the
        evaluation of ROC-AUC and Average Precision confidence margins can be
        rather slow for larger datasets.""",
    required=True,
    show_default=True,
    default=False,
    cls=ResourceOption,
)
@verbosity_option(logger=logger, expose_value=False)
def evaluate(
    predictions: pathlib.Path,
    output_folder: pathlib.Path,
    threshold: str | float,
    binning: str,
    plot: bool,
    credible_regions: bool,
    **_,  # ignored
) -> None:  # numpydoc ignore=PR01
    """Evaluate predictions (from a model) on a classification task."""
    import json
    import typing

    import matplotlib.backends.backend_pdf

    from ...engine.classify.evaluator import make_plots, make_table, run
    from ..utils import save_json_metadata, save_json_with_backup

    evaluation_file = output_folder / "evaluation.json"

    # register metadata
    save_json_metadata(
        output_file=evaluation_file.with_suffix(".meta.json"),
        predictions=str(predictions),
        output_folder=str(output_folder),
        threshold=threshold,
        binning=binning,
        plot=plot,
    )

    with predictions.open("r") as f:
        predict_data = json.load(f)

    if threshold in predict_data:
        # It is the name of a split on one of the two supported classifier
        # types. First run evaluation for reference dataset.
        from ...engine.classify.evaluator import maxf1_threshold

        use_threshold = maxf1_threshold(predict_data[threshold])
        logger.info(f"Setting --threshold={use_threshold:.5f}")

    else:
        # we try to convert it to float and complain if that is not possible
        try:
            use_threshold = float(threshold)
        except ValueError:
            raise click.BadParameter(
                f"""The value of --threshold=`{threshold}` does not match one
                of the database split names ({', '.join(predict_data.keys())})
                or can not be converted to a float. Check your input.""",
            )

    results: dict[str, dict[str, typing.Any]] = dict()
    for k, v in predict_data.items():
        logger.info(f"Computing performance on split `{k}`...")
        results[k] = run(
            name=k,
            predictions=v,
            binning=int(binning) if binning.isnumeric() else binning,
            threshold_a_priori=use_threshold,
            credible_regions=credible_regions,
        )

    # records full result analysis to a JSON file
    logger.info(f"Saving evaluation results at `{str(evaluation_file)}`...")
    save_json_with_backup(evaluation_file, results)

    # Produces and records table
    table = make_table(results, "rst")
    click.echo(table)

    output_table = evaluation_file.with_suffix(".rst")
    logger.info(f"Saving tabulated performance summary at `{str(output_table)}`...")
    output_table.parent.mkdir(parents=True, exist_ok=True)
    with output_table.open("w") as f:
        f.write(table)

    # Plots pre-calculated curves, if the user asked to do so.
    if plot:
        figure_path = evaluation_file.with_suffix(".pdf")
        logger.info(f"Saving evaluation figures at `{str(figure_path)}`...")
        with matplotlib.backends.backend_pdf.PdfPages(figure_path) as pdf:
            for fig in make_plots(results):
                pdf.savefig(fig)



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import click
from clapper.click import verbosity_option

from .logging import setup_cli_logger

logger = setup_cli_logger()


def _echo(left: str, right: str, color: str = "white") -> None:
    s = [
        click.style(left, bold=True),
        click.style(": ", bold=True),
        click.style(right, fg=color),
    ]
    click.echo("".join(s))


@click.command(
    epilog="""Examples:

1. Provide information about the current installation:

   .. code:: sh

      mednet info

""",
)
@verbosity_option(logger=logger, expose_value=False)
def info(
    **_,
) -> None:  # numpydoc ignore=PR01
    """Provide information about the current installation."""

    import typing

    from ..utils.rc import load_rc
    from .database import _list_raw_databases
    from .utils import execution_metadata

    m = execution_metadata()
    package_name = __name__.split(".")[0]

    _echo(f"{package_name} version", typing.cast(str, m["package-version"]), "green")
    _echo("platform", typing.cast(str, m["platform"]), "yellow")
    _echo(
        "accelerators",
        ", ".join(typing.cast(list[str], m["accelerators"])),
        "cyan",
    )

    rc = load_rc()

    if not rc.path.exists():
        _echo(f"{package_name} configuration", f"{str(rc.path)} [MISSING]", "white")
    else:
        _echo(f"{package_name} configuration file", f"{str(rc.path)}", "white")

    click.secho("databases:", bold=True)
    _list_raw_databases()

    click.secho("dependencies:", bold=True)
    python = typing.cast(dict[str, str], m["python"])
    _echo(
        "  - python",
        f"{python['version']} ({python['path']})",
        "white",
    )
    deps = typing.cast(dict[str, str], m["dependencies"])
    for name, version in deps.items():
        _echo(f"  - {name}", version, "white")


@click.command(
    epilog="""Examples:

1. Provide versio of the current installation:

   .. code:: sh

      mednet version

""",
)
@verbosity_option(logger=logger, expose_value=False)
def version(
    **_,
) -> None:  # numpydoc ignore=PR01
    """Provide version of the current installation."""

    import importlib.metadata

    package_name = __name__.split(".")[0]
    version = importlib.metadata.version(package_name)

    click.echo(version)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Feedforward shallow MLP for binary classification.

Simple feedforward MLP taking multiple inputs and generating a single
output (e.g. to predict active TB presence from radiological finding
estimates).
"""

from mednet.models.classify.mlp import MultiLayerPerceptron

model = MultiLayerPerceptron()



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Tests for the CPU/GPU resource monitor utilities."""

import os
import platform
import shutil
import time

import numpy
import pytest


def test_cpu_constants():
    from mednet.utils.resources import cpu_constants

    v = cpu_constants()
    assert "memory-total-GB/cpu" in v
    assert "number-of-cores/cpu" in v


def test_combined_monitor_cpu_only():
    from mednet.engine.device import DeviceManager
    from mednet.utils.resources import _CombinedMonitor

    monitor = _CombinedMonitor(
        device_type=DeviceManager("cpu").device_type,
        main_pid=os.getpid(),
    )

    log1 = monitor.log()

    assert isinstance(log1, dict)
    assert "timestamp" in log1
    assert "num-processes/cpu" in log1
    assert log1["num-processes/cpu"] >= 1

    log2 = monitor.log()

    assert isinstance(log2, dict)
    assert "timestamp" in log2
    assert log1["timestamp"] != log2["timestamp"]


@pytest.mark.skipif(
    (platform.system().lower() != "darwin" or platform.processor() != "arm"),
    reason="Requires macOS on Apple silicon to run",
)
def test_mps_constants():
    from mednet.utils.resources import mps_constants

    v = mps_constants()
    assert "apple-processor-model" in v
    assert "number-of-cores/gpu" in v


@pytest.mark.skipif(
    (platform.system().lower() != "darwin" or platform.processor() != "arm"),
    reason="Requires macOS on Apple silicon to run",
)
def test_combined_monitor_macos_gpu():
    from mednet.engine.device import DeviceManager
    from mednet.utils.resources import _CombinedMonitor

    monitor = _CombinedMonitor(
        device_type=DeviceManager("mps").device_type,
        main_pid=os.getpid(),
    )

    log1 = monitor.log()

    assert isinstance(log1, dict)
    assert "timestamp" in log1
    assert "num-processes/cpu" in log1
    assert log1["num-processes/cpu"] >= 1
    assert "percent-usage/gpu" in log1
    assert "frequency-MHz/gpu" in log1

    log2 = monitor.log()

    assert isinstance(log2, dict)
    assert "timestamp" in log2
    assert "percent-usage/gpu" in log2
    assert "frequency-MHz/gpu" in log2
    assert log1["timestamp"] != log2["timestamp"]


@pytest.mark.skipif(
    shutil.which("nvidia-smi") is None, reason="Requires nvidia-smi to run"
)
def test_cuda_constants():
    from mednet.utils.resources import cuda_constants

    v = cuda_constants()
    assert "driver-version/gpu" in v
    assert "total-memory-GB/gpu" in v


@pytest.mark.skipif(
    shutil.which("nvidia-smi") is None, reason="Requires nvidia-smi to run"
)
def test_combined_monitor_nvidia_gpu():
    from mednet.engine.device import DeviceManager
    from mednet.utils.resources import _CombinedMonitor

    monitor = _CombinedMonitor(
        device_type=DeviceManager("cuda").device_type,
        main_pid=os.getpid(),
    )

    log1 = monitor.log()

    assert isinstance(log1, dict)
    assert "timestamp" in log1
    assert "num-processes/cpu" in log1
    assert log1["num-processes/cpu"] >= 1
    assert "percent-usage/gpu" in log1
    assert "memory-used-GB/gpu" in log1
    assert "memory-percent/gpu" in log1

    log2 = monitor.log()

    assert isinstance(log2, dict)
    assert "timestamp" in log2
    assert "percent-usage/gpu" in log2
    assert "memory-used-GB/gpu" in log2
    assert "memory-percent/gpu" in log2
    assert log1["timestamp"] != log2["timestamp"]


def test_aggregation():
    from mednet.engine.device import DeviceManager
    from mednet.utils.resources import _CombinedMonitor, aggregate

    monitor = _CombinedMonitor(
        device_type=DeviceManager("cpu").device_type,
        main_pid=os.getpid(),
    )

    logs = []

    for _ in range(5):
        time.sleep(0.2)
        logs.append(monitor.log())

    agg = aggregate(logs)
    assert "cpu-monitoring-samples" in agg
    assert agg["cpu-monitoring-samples"] == 5

    assert "timestamp" in agg
    assert agg["timestamp"] == logs[-1]["timestamp"] - logs[0]["timestamp"]

    assert "num-processes/cpu" in agg
    assert agg["num-processes/cpu"] == max([k["num-processes/cpu"] for k in logs])

    assert "num-open-files/cpu" in agg
    assert agg["num-open-files/cpu"] == max([k["num-open-files/cpu"] for k in logs])

    assert "percent-usage/cpu" in agg
    assert numpy.isclose(
        agg["percent-usage/cpu"],
        sum([k["percent-usage/cpu"] for k in logs]) / agg["cpu-monitoring-samples"],
    )


def test_mp_cpu_monitoring():
    # Checks a "normal" workflow, where the monitoring interval is smaller than
    # the total work time

    from mednet.engine.device import DeviceManager
    from mednet.utils.resources import ResourceMonitor, aggregate

    rm = ResourceMonitor(
        interval=0.2,
        device_type=DeviceManager("cpu").device_type,
        main_pid=os.getpid(),
    )

    with rm:
        time.sleep(0.5)
        logs = rm.checkpoint()
        assert len(logs) >= 2
        agg = aggregate(logs)
        assert agg["cpu-monitoring-samples"] >= 2


def test_mp_cpu_monitoring_short_processing():
    # Checks we can get at least 1 monitoring sample even if the processing is
    # super short

    from mednet.engine.device import DeviceManager
    from mednet.utils.resources import ResourceMonitor, aggregate

    rm = ResourceMonitor(
        interval=0.5,
        device_type=DeviceManager("cpu").device_type,
        main_pid=os.getpid(),
    )

    with rm:
        time.sleep(0.2)  # shorter than interval
        logs = rm.checkpoint()
        assert len(logs) == 1
        agg = aggregate(logs)
        assert agg["cpu-monitoring-samples"] == 1


@pytest.mark.skipif(
    (platform.system().lower() != "darwin" or platform.processor() != "arm"),
    reason="Requires macOS on Apple silicon to run",
)
def test_mp_macos_gpu_monitoring():
    # Checks a "normal" workflow, where the monitoring interval is smaller than
    # the total work time

    from mednet.engine.device import DeviceManager
    from mednet.utils.resources import ResourceMonitor, aggregate

    rm = ResourceMonitor(
        interval=1.5,
        device_type=DeviceManager("mps").device_type,
        main_pid=os.getpid(),
    )

    with rm:
        time.sleep(3.0)
        logs = rm.checkpoint()
        assert len(logs) >= 1
        agg = aggregate(logs)
        assert agg["cpu-monitoring-samples"] >= 1


@pytest.mark.skipif(
    (platform.system().lower() != "darwin" or platform.processor() != "arm"),
    reason="Requires macOS on Apple silicon to run",
)
def test_mp_macos_gpu_monitoring_short_processing():
    # Checks we can get at least 1 monitoring sample even if the processing is
    # shorter.  In this check we execute an external utility which may
    # delay obtaining samples.

    from mednet.engine.device import DeviceManager
    from mednet.utils.resources import ResourceMonitor, aggregate

    rm = ResourceMonitor(
        interval=1.5,  # on my mac, this measurements take ~0.9s
        device_type=DeviceManager("mps").device_type,
        main_pid=os.getpid(),
    )

    with rm:
        time.sleep(1)  # shorter than interval
        logs = rm.checkpoint()
        assert len(logs) == 1
        agg = aggregate(logs)
        assert agg["cpu-monitoring-samples"] == 1



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib
import time
from importlib.metadata import distribution

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = "1.3"

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.todo",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.graphviz",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "auto_intersphinx",
    "sphinx_autodoc_typehints",
    "sphinx_copybutton",
    "sphinx_inline_tabs",
    "sphinx_click",
    "sphinxcontrib.bibtex",
]

# Be picky about warnings
nitpicky = True

# Ignores stuff we can't easily resolve on other project's sphinx manuals
nitpick_ignore = []

# Allows the user to override warnings from a separate file
nitpick_path = pathlib.Path("nitpick-exceptions.txt")
if nitpick_path.exists():
    for line in nitpick_path.open():
        if line.strip() == "" or line.startswith("#"):
            continue
        dtype, target = line.split(None, 1)
        target = target.strip()
        nitpick_ignore.append((dtype, target))

# Always includes todos
todo_include_todos = True

# Generates auto-summary automatically
autosummary_generate = True

# Create numbers on figures with captions
numfig = True

# If we are on OSX, the 'dvipng' path maybe different
dvipng_osx = pathlib.Path("/Library/TeX/texbin/dvipng")
if dvipng_osx.exists():
    pngmath_dvipng = str(dvipng_osx)

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = {".rst": "restructuredtext"}

# The main toctree document.
main_doc = "index"

# General information about the project.
project = "mednet"
package = distribution(project)

copyright = f"{time.strftime('%Y')}, Idiap Research Institute"  # noqa

# The short X.Y version.
version = package.version
# The full version, including alpha/beta/rc tags.
release = version

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["links.rst"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"
pygments_dark_style = "monokai"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# Some variables which are useful for generated material
project_variable = project.replace(".", "_")
short_description = package.metadata["Summary"]
owner = ["Idiap Research Institute"]

# Graphviz output
graphviz_output_format = "svg"

# BibTex extension
bibtex_bibfiles = ["references.bib"]

# -- Options for HTML output ---------------------------------------------------

html_theme = "furo"

html_theme_options = {
    "source_edit_link": f"https://gitlab.idiap.ch/medai/software/{project}/-/edit/main/doc/{{filename}}",
    "source_view_link": f"https://gitlab.idiap.ch/medai/software/{project}/-/tree/main/doc/{{filename}}",
}

html_title = f"{project} {release}"

# -- Post configuration --------------------------------------------------------

# Default processing flags for sphinx
autodoc_member_order = "bysource"
autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "show-inheritance": True,
}

auto_intersphinx_packages = [
    "matplotlib",
    "numpy",
    "pillow",
    "psutil",
    "scipy",
    "torch",
    "torchvision",
    "lightning",
    "tensorboardx",
    "clapper",
    ("python", "3"),
    "python-gitlab",
    "torchio",
    "credible",
    "click",
]
auto_intersphinx_catalog = "catalog.json"

# Add our private index (for extras and fixes)
intersphinx_mapping = dict(extras=("https://mednet.rtfd.io", "extras.inv"))



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Horizontal flipping (p=50%) followed by simple affine augmentations."""

from .affine import augmentations as _affine
from .hflip import augmentations as _hflip

augmentations = _hflip + _affine



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""`AlexNet network architecture <alexnet-pytorch_>`_, from :cite:p:`krizhevsky_imagenet_2017`."""

import logging
import typing

import torch
import torch.nn
import torch.optim.optimizer
import torch.utils.data
import torchvision.models as models

from ...data.typing import TransformSequence
from ..typing import Checkpoint
from .model import Model

logger = logging.getLogger(__name__)


class Alexnet(Model):
    """`AlexNet network architecture <alexnet-pytorch_>`_ model, from :cite:p:`krizhevsky_imagenet_2017`.

    Note: only usable with a normalized dataset

    Parameters
    ----------
    loss_type
        The loss to be used for training and evaluation.

        .. warning::

           The loss should be set to always return batch averages (as opposed
           to the batch sum), as our logging system expects it so.
    loss_arguments
        Arguments to the loss.
    optimizer_type
        The type of optimizer to use for training.
    optimizer_arguments
        Arguments to the optimizer after ``params``.
    scheduler_type
        The type of scheduler to use for training.
    scheduler_arguments
        Arguments to the scheduler after ``params``.
    model_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.
    augmentation_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.
    pretrained
        If set to True, loads pretrained model weights during initialization,
        else trains a new model.
    num_classes
        Number of outputs (classes) for this model.
    """

    def __init__(
        self,
        loss_type: type[torch.nn.Module] = torch.nn.BCEWithLogitsLoss,
        loss_arguments: dict[str, typing.Any] | None = None,
        optimizer_type: type[torch.optim.Optimizer] = torch.optim.Adam,
        optimizer_arguments: dict[str, typing.Any] | None = None,
        scheduler_type: type[torch.optim.lr_scheduler.LRScheduler] | None = None,
        scheduler_arguments: dict[str, typing.Any] | None = None,
        model_transforms: TransformSequence | None = None,
        augmentation_transforms: TransformSequence | None = None,
        pretrained: bool = False,
        num_classes: int = 1,
    ):
        super().__init__(
            name="alexnet",
            loss_type=loss_type,
            loss_arguments=loss_arguments,
            optimizer_type=optimizer_type,
            optimizer_arguments=optimizer_arguments,
            scheduler_type=scheduler_type,
            scheduler_arguments=scheduler_arguments,
            model_transforms=model_transforms,
            augmentation_transforms=augmentation_transforms,
            num_classes=num_classes,
        )

        self.pretrained = pretrained

        # Load pretrained model
        weights = None
        if self.pretrained:
            from ..normalizer import make_imagenet_normalizer

            Model.normalizer.fset(self, make_imagenet_normalizer())  # type: ignore[attr-defined]

            logger.info(f"Loading pretrained `{self.name}` model weights")
            weights = models.AlexNet_Weights.DEFAULT

        self.model = models.alexnet(weights=weights)

        self.model.classifier[4] = torch.nn.Linear(
            in_features=self.model.classifier[1].out_features, out_features=512
        )
        self.model.classifier[6] = torch.nn.Linear(
            in_features=self.model.classifier[4].out_features,
            out_features=self.num_classes,
        )

    @Model.num_classes.setter  # type: ignore[attr-defined]
    def num_classes(self, v: int) -> None:
        if self.num_classes != v:
            logger.info(
                f"Resetting `{self.name}` output classifier layer weights due "
                f"to a change in output size ({self.num_classes} -> {v})"
            )
            self.model.classifier[6] = torch.nn.Linear(
                in_features=self.model.classifier[4].out_features, out_features=v
            )
            self._num_classes = v

    def on_load_checkpoint(self, checkpoint: Checkpoint) -> None:
        # reset number of output classes if need be
        self.num_classes = checkpoint["state_dict"]["model.classifier.bias"].shape[0]

        # perform routine checkpoint loading
        super().on_load_checkpoint(checkpoint)

    def forward(self, x):
        """Forward the input tensor through the network, producing a prediction.

        Parameters
        ----------
        x
            The tensor input to be forwarded.

        Returns
        -------
            The prediction, as a tensor.
        """
        x = self.normalizer(x)
        return self.model(x)



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""MobileNetV2 U-Net model for image segmentation.

The MobileNetV2 architecture is based on an inverted residual structure where
the input and output of the residual block are thin bottleneck layers opposite
to traditional residual models which use expanded representations in the input
an MobileNetV2 uses lightweight depthwise convolutions to filter features in
the intermediate expansion layer.  This model implements a MobileNetV2 U-Net
model, henceforth named M2U-Net, combining the strenghts of U-Net for medical
segmentation applications and the speed of MobileNetV2 networks.

References: :cite:p:`sandler_mobilenetv2_2018`, :cite:p:`navab_u-net_2015`
"""

import torch.optim
import torchvision.transforms
import torchvision.transforms.v2

import mednet.models.losses
import mednet.models.segment.m2unet
import mednet.models.transforms

model = mednet.models.segment.m2unet.M2Unet(
    loss_type=mednet.models.losses.SoftJaccardAndBCEWithLogitsLoss,
    loss_arguments=dict(alpha=0.7),  # 0.7 BCE + 0.3 Jaccard
    optimizer_type=torch.optim.Adam,
    optimizer_arguments=dict(lr=0.01),
    model_transforms=[
        mednet.models.transforms.SquareCenterPad(),
        torchvision.transforms.v2.Resize(512, antialias=True),
        torchvision.transforms.v2.RGB(),
    ],
    pretrained=False,
)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import pathlib
import typing

import h5py
import lightning.pytorch
import torch.utils.data
import tqdm

logger = logging.getLogger(__name__)


def run(
    datamodule: lightning.pytorch.LightningDataModule,
    output_folder: pathlib.Path,
) -> (
    dict[str, list[tuple[str, str]]]
    | list[list[tuple[str, str]]]
    | list[tuple[str, str]]
    | None
):
    """Dump annotations from input datamodule.

    Parameters
    ----------
    datamodule
        The lightning DataModule to extract annotations from.
    output_folder
        Folder where to store HDF5 representations of annotations.

    Returns
    -------
        A JSON-able representation of sample data stored at ``output_folder``.
        For every split (dataloader), a list of samples in the form
        ``[sample-name, hdf5-path]`` is returned.  In the cases where the
        ``predict_dataloader()`` returns a single loader, we then return a
        list.  A dictionary is returned in case ``predict_dataloader()`` also
        returns a dictionary.

    Raises
    ------
    TypeError
        If the DataModule's ``predict_dataloader()`` method does not return any
        of the types described above.
    """

    def _write_sample(
        sample: typing.Any, output_folder: pathlib.Path
    ) -> tuple[str, str]:
        """Write a single sample target to an HDF5 file.

        Parameters
        ----------
        sample
            A segmentation sample as output by a dataloader.
        output_folder
            Path leading to a folder where to store dumped annotations.

        Returns
        -------
            A tuple which contains the sample path and the destination
            directory where the HDF5 file was saved.
        """
        name = sample["name"][0]
        stem = pathlib.Path(name).with_suffix(".hdf5")
        dest = output_folder / stem
        tqdm.tqdm.write(f"`{name}` -> `{str(dest)}`")
        dest.parent.mkdir(parents=True, exist_ok=True)
        with h5py.File(dest, "w") as f:
            f.create_dataset(
                "image",
                data=sample["image"][0].cpu().numpy(),
                compression="gzip",
                compression_opts=9,
            )
            f.create_dataset(
                "target",
                data=(sample["target"][0].squeeze(0).cpu().numpy() > 0.5),
                compression="gzip",
                compression_opts=9,
            )
            f.create_dataset(
                "mask",
                data=(sample["mask"][0].squeeze(0).cpu().numpy() > 0.5),
                compression="gzip",
                compression_opts=9,
            )
        return (name, str(stem))

    dataloaders = datamodule.predict_dataloader()

    if isinstance(dataloaders, torch.utils.data.DataLoader):
        logger.info("Dump annotations from a single dataloader...")
        return [_write_sample(k, output_folder) for k in tqdm.tqdm(dataloaders)]

    if isinstance(dataloaders, list):
        retval_list = []
        for k, dataloader in enumerate(dataloaders):
            logger.info(f"Dumping annotations from split `{k}`...")
            retval_list.append(
                [_write_sample(k, output_folder) for k in tqdm.tqdm(dataloader)]
            )
        return retval_list

    if isinstance(dataloaders, dict):
        retval_dict = {}
        for name, dataloader in dataloaders.items():
            logger.info(f"Dumping annotations from split `{name}`...")
            retval_dict[name] = [
                _write_sample(k, output_folder) for k in tqdm.tqdm(dataloader)
            ]
        return retval_dict

    if dataloaders is None:
        logger.warning("Datamodule did not return any prediction dataloaders!")
        return None

    # if you get to this point, then the user is returning something that is
    # not supported - complain!
    raise TypeError(
        f"Datamodule returned strangely typed prediction "
        f"dataloaders: `{type(dataloaders)}` - Please write code "
        f"to support this use-case.",
    )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import click
from clapper.click import AliasedGroup, verbosity_option

from .logging import setup_cli_logger

logger = setup_cli_logger()


@click.group(cls=AliasedGroup)
def config():
    """Command for listing, describing and copying configuration resources."""
    pass


@config.command(
    name="list",
    epilog="""Examples:

\b
  1. Lists all configuration resources installed:

     .. code:: sh

        mednet config list


\b
  2. Lists all configuration resources and their descriptions (notice this may
     be slow as it needs to load all modules once):

     .. code:: sh

        mednet config list -v

""",
)
@verbosity_option(logger=logger, expose_value=False)
@click.pass_context
def list_(ctx, **_) -> None:  # numpydoc ignore=PR01
    """List configuration files installed."""

    import importlib.metadata
    import inspect
    import typing

    entry_points = importlib.metadata.entry_points().select(group="mednet.config")
    entry_point_dict = {k.name: k for k in entry_points}

    # all potential modules with configuration resources
    modules = {k.module.rsplit(".", 1)[0] for k in entry_point_dict.values()}

    # sort data entries by originating module
    entry_points_by_module: dict[str, dict[str, typing.Any]] = {}
    for k in modules:
        entry_points_by_module[k] = {}
        for name, ep in entry_point_dict.items():
            if ep.module.rsplit(".", 1)[0] == k:
                entry_points_by_module[k][name] = ep

    for config_type in sorted(entry_points_by_module):
        # calculates the longest config name so we offset the printing
        longest_name_length = max(
            len(k) for k in entry_points_by_module[config_type].keys()
        )

        # set-up printing options
        print_string = "  %%-%ds   %%s" % (longest_name_length,)
        # 79 - 4 spaces = 75 (see string above)
        description_leftover = 75 - longest_name_length

        click.echo(f"module: {config_type}")
        for name in sorted(entry_points_by_module[config_type]):
            ep = entry_point_dict[name]

            if ctx.meta.get("verbose", 0) >= 1:
                module = ep.load()
                doc = inspect.getdoc(module)
                if doc is not None:
                    summary = doc.split("\n\n")[0]
                else:
                    summary = "<DOCSTRING NOT AVAILABLE>"
            else:
                summary = ""

            summary = (
                (summary[: (description_leftover - 3)] + "...")
                if len(summary) > (description_leftover - 3)
                else summary
            )

            click.echo(print_string % (name, summary))


@config.command(
    epilog="""Examples:

\b
  1. Describe the Montgomery dataset configuration:

     .. code:: sh

        mednet config describe montgomery


\b
  2. Describe the Montgomery dataset configuration and lists its
     contents:

     .. code:: sh

        mednet config describe montgomery -v

""",
)
@click.argument(
    "name",
    required=True,
    nargs=-1,
)
@verbosity_option(logger=logger, expose_value=False)
@click.pass_context
def describe(ctx, name, **_) -> None:  # numpydoc ignore=PR01
    """Describe a specific configuration file."""

    import importlib.metadata
    import inspect
    import pathlib

    entry_points = importlib.metadata.entry_points().select(group="mednet.config")
    entry_point_dict = {k.name: k for k in entry_points}

    for k in name:
        if k not in entry_point_dict:
            logger.error("Cannot find configuration resource '%s'", k)
            continue
        ep = entry_point_dict[k]
        click.echo(f"Configuration: {ep.name}")
        click.echo(f"Python Module: {ep.module}")
        click.echo("")
        mod = ep.load()

        if ctx.meta.get("verbose", 0) >= 1:
            fname = inspect.getfile(mod)
            click.echo("Contents:")
            with pathlib.Path(fname).open() as f:
                click.echo(f.read())
        else:  # only output documentation
            click.echo("Documentation:")
            click.echo(inspect.getdoc(mod))


@config.command(
    epilog="""Examples:

\b
  1. Make a copy of one of the stock configuration files locally, so it can be
     adapted:

     .. code:: sh

        $ mednet config copy montgomery -vvv newdataset.py

""",
)
@click.argument(
    "source",
    required=True,
    nargs=1,
)
@click.argument(
    "destination",
    required=True,
    nargs=1,
)
@verbosity_option(logger=logger, expose_value=False)
def copy(source, destination) -> None:  # numpydoc ignore=PR01
    """Copy a specific configuration resource so it can be modified locally."""

    import importlib.metadata
    import inspect
    import shutil

    entry_points = importlib.metadata.entry_points().select(group="mednet.config")
    entry_point_dict = {k.name: k for k in entry_points}

    if source not in entry_point_dict:
        logger.error("Cannot find configuration resource '%s'", source)
        return

    ep = entry_point_dict[source]
    mod = ep.load()
    src_name = inspect.getfile(mod)
    logger.info(f"cp {src_name} -> {destination}")
    shutil.copyfile(src_name, destination)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib

import click
from clapper.click import ResourceOption, verbosity_option

from .click import ConfigCommand
from .logging import setup_cli_logger

logger = setup_cli_logger()


@click.command(
    entry_point_group="mednet.config",
    cls=ConfigCommand,
    epilog="""Examples:

1. Upload an existing experiment result from a path it resides on (with a default experiment name as {model-name}_{database-name} and a default run name as {date-time}):

   .. code:: sh

      mednet upload --experiment-folder=/path/to/results

2. Upload an existing experiment result with an experiment name:

   .. code:: sh

      mednet upload --experiment-folder=/path/to/results --experiment-name=exp-pasa_mc

3. Upload an existing experiment result with a run name:

   .. code:: sh

      mednet upload --experiment-folder=/path/to/results --run-name=run-1

4. Upload an existing experiment result with defining a size limit of 20MB for each file:

   .. code:: sh

      mednet upload --experiment-folder=/path/to/results --upload-limit-mb=20

""",
)
@click.option(
    "--project-path",
    "-p",
    help="Path to the project where to upload model entries",
    required=True,
    type=str,
    default="medai/software/mednet",
    show_default=True,
    cls=ResourceOption,
)
@click.option(
    "--experiment-folder",
    "-e",
    help="Directory in which to upload results from",
    required=True,
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        path_type=pathlib.Path,
    ),
    default="results",
    show_default=True,
    cls=ResourceOption,
)
@click.option(
    "--experiment-name",
    "-n",
    help='A string indicating the experiment name (e.g. "exp-pasa-mc" or "exp-densenet-mc-ch")',
    required=True,
    cls=ResourceOption,
)
@click.option(
    "--run-name",
    "-r",
    help='A string indicating the run name (e.g. "run-1")',
    required=True,
    default="",
    cls=ResourceOption,
)
@click.option(
    "--upload-limit-mb",
    "-l",
    help="Maximim upload size in MB (set to 0 for no limit).",
    show_default=True,
    required=True,
    default=10,
    type=click.IntRange(min=0),
    cls=ResourceOption,
)
@verbosity_option(logger=logger, expose_value=False)
def upload(
    project_path: str,
    experiment_folder: pathlib.Path,
    experiment_name: str,
    run_name: str,
    upload_limit_mb: int,
    **_,  # ignored
) -> None:  # numpydoc ignore=PR01
    """Upload results from a classification experiment folder to GitLab's MLFlow server."""

    from ..engine.uploader import run

    # further metrics to be displayed
    metrics = [
        "threshold",
        "precision",
        "recall",
        "f1",
        "average_precision",
        "specificity",
        "roc_auc",
        "accuracy",
    ]

    run(
        project_path=project_path,
        experiment_folder=experiment_folder,
        experiment_name=experiment_name,
        run_name=run_name,
        metrics=metrics,
        upload_limit_mb=upload_limit_mb,
    )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import click
from clapper.click import AliasedGroup, verbosity_option

from .logging import setup_cli_logger

logger = setup_cli_logger()


def _get_raw_databases() -> dict[str, dict[str, str | list]]:
    """Return a list of all supported (raw) databases.

    Returns
    -------
    dict[str, dict[str, str]]
        Dictionary where keys are database names, and values are dictionaries
        containing two string keys:

        * ``module``: the full Pythonic module name (e.g.
          ``mednet.data.classify.montgomery``).
        * ``datadir``: points to the user-configured data directory for the
          current database, if set, or ``None`` otherwise.
    """

    import importlib
    import pkgutil
    import typing

    from ..data import classify as classify_data
    from ..data import detect as detect_data
    from ..data import segment as segment_data
    from ..utils.rc import load_rc

    user_configuration = load_rc()

    retval: dict[str, dict[str, typing.Any]] = {}
    for data in (classify_data, segment_data, detect_data):
        for k in pkgutil.iter_modules(data.__path__):
            assert data.__package__ is not None
            module = importlib.import_module(data.__package__ + "." + k.name)
            datadir = getattr(module, "CONFIGURATION_KEY_DATADIR", None)
            if datadir is None:
                continue

            retval.setdefault(
                k.name,
                dict(module=[], datadir=user_configuration.get(datadir)),
            )["module"].append(module.__name__.rsplit(".", 1)[0])

    return retval


def _list_raw_databases():
    """List raw databases to a string representation."""

    def _echo(left: str, right: str, color: str = "white") -> None:
        s = [
            click.style(left, bold=True),
            click.style(": ", bold=True),
            click.style(right, fg=color),
        ]
        click.echo("".join(s))

    for k, v in _get_raw_databases().items():
        if "datadir" not in v:
            # this database does not have a "datadir"
            continue

        if v["datadir"] is not None:
            _echo(f"  - {k} ({', '.join(v['module'])})", f"{v['datadir']}", "green")
        else:
            _echo(f"  - {k} ({', '.join(v['module'])})", "NOT installed", "red")


@click.group(cls=AliasedGroup)
def database() -> None:
    """Command for listing and verifying databases installed."""
    pass


@database.command(
    name="list",
    epilog="""Examples:

\b
    1. To install a database, set up its data directory ("datadir").  For
       example, to setup access to Montgomery files you downloaded locally at
       the directory "/path/to/montgomery/files", edit the RC file (typically
       ``$HOME/.config/mednet.toml``), and add a line like the following:

       .. code:: toml

          [datadir]
          montgomery = "/path/to/montgomery/files"

       .. note::

          This setting **is** case-sensitive.

\b
    2. List all raw databases supported (and configured):

       .. code:: sh

          $ mednet database list

""",
)
@verbosity_option(logger=logger, expose_value=False)
def list_():
    """List all supported and configured databases."""

    _list_raw_databases()


@database.command(
    epilog="""Examples:

    1. Check if all files from the config split 'montgomery-f0' of the Montgomery
       database can be loaded:

       .. code:: sh

          mednet database check -vv montgomery-f0

""",
)
@click.argument(
    "entrypoint",
    nargs=1,
)
@click.option(
    "--limit",
    "-l",
    help="Limit check to the first N samples in each split in the "
    "configuration, making the check sensibly faster. Set it to "
    "zero (default) to check everything.",
    required=True,
    type=click.IntRange(0),
    default=0,
    show_default=True,
)
@verbosity_option(logger=logger, expose_value=False)
def check(entrypoint, limit):  # numpydoc ignore=PR01
    """Check file access on a database configuration split."""
    import importlib.metadata
    import sys

    click.secho(f"Checking database split config `{entrypoint}`...", fg="yellow")
    try:
        module = importlib.metadata.entry_points(group="mednet.config")[
            entrypoint
        ].module
    except KeyError:
        raise Exception(f"Could not find database split config `{entrypoint}`")

    datamodule = importlib.import_module(module).datamodule

    datamodule.model_transforms = []  # should be done before setup()
    datamodule.batch_size = 1  # ensure one sample is loaded at a time
    datamodule.setup("predict")  # sets up all datasets

    loaders = datamodule.predict_dataloader()

    errors = 0
    for k, loader in loaders.items():
        if limit == 0:
            click.secho(
                f"Checking all {len(loader)} samples of split `{k}` at config "
                f"`{entrypoint}`...",
                fg="yellow",
            )
            loader_limit = sys.maxsize
        else:
            click.secho(
                f"Checking first {limit} samples of dataset "
                f"`{k}` at config `{entrypoint}`...",
                fg="yellow",
            )
            loader_limit = limit

        # the for loop will trigger raw data loading (ie. user code), protect it
        i = 0
        try:
            for i, batch in enumerate(loader):
                if loader_limit == 0:
                    break
                logger.info(
                    f"{batch['name'][0]}: "
                    f"{[s for s in batch['image'][0].shape]}@{batch['image'][0].dtype}",
                )
                loader_limit -= 1
        except Exception:
            logger.exception(f"Unable to load sample {i} at split {k}")
            errors += 1

    if not errors:
        click.secho(
            f"OK! No errors were reported for database entrypoint `{entrypoint}`.",
            fg="green",
        )
    else:
        click.secho(
            f"Found {errors} errors loading DataModule `{entrypoint}`.",
            fg="red",
        )


# Add "preprocess" subcommand
database.add_command(
    getattr(
        __import__("importlib").import_module("..preprocess", package=__name__),
        "preprocess",
    )
)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Engine and functions for human interpretability analysis."""

import logging
import pathlib
import typing

import lightning.pytorch
import numpy
import numpy.typing
from torchvision import tv_tensors
from tqdm import tqdm

logger = logging.getLogger(__name__)

SaliencyMap: typing.TypeAlias = (
    typing.Sequence[typing.Sequence[float]] | numpy.typing.NDArray[numpy.double]
)
BinaryMask: typing.TypeAlias = numpy.typing.NDArray[numpy.bool_]


def _compute_avg_saliency_focus(
    saliency_map: SaliencyMap,
    gt_mask: BinaryMask,
) -> float:
    """Integrate the saliency map over the ground-truth boxes and normalizes by
    total bounding-box area.

    This function will integrate (sum) the value of the saliency map over the
    ground-truth bounding boxes and normalize it by the total area covered by
    all ground-truth bounding boxes.

    Parameters
    ----------
    saliency_map
        A real-valued saliency-map that conveys regions used for
        classification in the original sample.
    gt_mask
        Ground-truth mask containing the bounding boxes of the ground-truth
        drawn as ``True`` values.

    Returns
    -------
    float
        A single floating-point number representing the Average saliency focus.
    """

    area = gt_mask.sum()
    if area == 0:
        return 0.0

    return numpy.sum(saliency_map * gt_mask) / area


def _compute_proportional_energy(
    saliency_map: SaliencyMap,
    gt_mask: BinaryMask,
) -> float:
    """Calculate how much activation lies within the ground truth boxes
    compared to the total sum of the activations (integral).

    Parameters
    ----------
    saliency_map
        A real-valued saliency-map that conveys regions used for
        classification in the original sample.
    gt_mask
        Ground-truth mask containing the bounding boxes of the ground-truth
        drawn as ``True`` values.

    Returns
    -------
    float
        A single floating-point number representing the proportional energy.
    """

    denominator = numpy.sum(saliency_map)

    if denominator == 0.0:
        return 0.0

    return float(numpy.sum(saliency_map * gt_mask) / denominator)  # type: ignore


def _compute_binary_mask(
    gt_bboxes: tv_tensors.BoundingBoxes,
    saliency_map: SaliencyMap,
) -> BinaryMask:
    """Compute a binary mask for the saliency map using BoundingBoxes.

    The binary_mask will be ON/True where the gt boxes are located.

    Parameters
    ----------
    gt_bboxes
        Ground-truth bounding boxes in the format ``(x, y, width,
        height)``.
    saliency_map
        A real-valued saliency-map that conveys regions used for
        classification in the original sample.

    Returns
    -------
    BinaryMask
        A numpy array of the same size as saliency_map with
        the value False everywhere except at the positions inside
        the bounding boxes, which will be True.
    """

    binary_mask = numpy.zeros_like(saliency_map, dtype=numpy.bool_)
    if gt_bboxes.format != tv_tensors.BoundingBoxFormat.XYXY:
        raise ValueError(
            f"Only boundingBoxes of format xyxy are supported. Got {gt_bboxes.format}."
        )

    for bbox in gt_bboxes:
        binary_mask[
            bbox.data[1] : bbox.data[1] + (bbox.data[3] - bbox.data[1]),
            bbox.data[0] : bbox.data[0] + (bbox.data[2] - bbox.data[0]),
        ] = True
    return binary_mask


def _process_sample(
    gt_bboxes: tv_tensors.BoundingBoxes,
    saliency_map: SaliencyMap,
) -> tuple[float, float]:
    """Calculate the metrics for a single sample.

    Parameters
    ----------
    gt_bboxes
        A list of ground-truth bounding boxes.
    saliency_map
        A real-valued saliency-map that conveys regions used for
        classification in the original sample.

    Returns
    -------
    tuple[float, float]
        A tuple containing the following values:

        * Proportional energy
        * Average saliency focus
    """

    binary_mask = _compute_binary_mask(gt_bboxes, saliency_map)

    return (
        _compute_proportional_energy(saliency_map, binary_mask),
        _compute_avg_saliency_focus(saliency_map, binary_mask),
    )


def run(
    input_folder: pathlib.Path,
    target_label: int,
    datamodule: lightning.pytorch.LightningDataModule,
    only_dataset: str | None,
) -> dict[str, list[typing.Any]]:
    """Compute the proportional energy and average saliency focus for a given
    target label in a DataModule.

    Parameters
    ----------
    input_folder
        Directory in which the saliency maps are stored for a specific
        visualization type.
    target_label
        The label to target for evaluating interpretability metrics. Samples
        contining any other label are ignored.
    datamodule
        The lightning DataModule to iterate on.
    only_dataset
        If set, will only run this code for the named dataset on the provided
        datamodule, skipping any other datasets.

    Returns
    -------
        A dictionary where keys are dataset names in the provided DataModule,
        and values are lists containing sample information alongside metrics
        calculated:

        * Sample name (str)
        * Sample target class (int)
        * Proportional energy (float)
        * Average saliency focus (float)
    """

    retval: dict[str, list[typing.Any]] = {}

    # TODO: This loads the images from the dataset, but they are not useful at
    # this point.  Possibly using the contents of ``datamodule.splits`` can
    # substantially speed this up.
    for dataset_name, dataset_loader in datamodule.predict_dataloader().items():
        if only_dataset is not None and dataset_name != only_dataset:
            logger.warning(
                f"Skipping processing for dataset `{dataset_name}` following user request..."
            )
            continue

        logger.info(
            f"Estimating interpretability metrics for dataset `{dataset_name}`...",
        )
        retval[dataset_name] = []

        for sample in tqdm(
            dataset_loader,
            desc="batches",
            leave=False,
            disable=None,
        ):
            name = str(sample["name"][0])
            label = int(sample["target"].item())

            if label != target_label:
                # we add the entry for dataset completeness, but do not treat
                # it
                retval[dataset_name].append([name, label])
                continue

            # TODO: This is very specific to the TBX11k system for labelling
            # regions of interest.  We need to abstract from this to support more
            # datasets and other ways to annotate.
            bboxes: tv_tensors.BoundingBoxes | None = sample.get("bboxes", None)

            if bboxes is None:
                logger.warning(
                    f"Sample `{name}` does not contain bounding-box information. "
                    f"No localization metrics can be calculated in this case. "
                    f"Skipping...",
                )
                # we add the entry for dataset completeness
                retval[dataset_name].append([name, label])
                continue

            # we fully process this entry
            retval[dataset_name].append(
                [
                    name,
                    label,
                    *_process_sample(
                        bboxes[0],
                        numpy.load(
                            input_folder / pathlib.Path(name).with_suffix(".npy"),
                        ),
                    ),
                ],
            )

    return retval



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Test code for datasets."""

import typing

import pytest
import torch
from torchvision import tv_tensors

import mednet.data.typing

_NUM_SAMPLES = 1000
_raw_dataset = [(f"sample-{k:3d}", k, f"metadata-{k:3d}") for k in range(_NUM_SAMPLES)]


class _RawDataLoader(mednet.data.typing.RawDataLoader):
    def sample(
        self, sample: tuple[str, int, typing.Any | None]
    ) -> mednet.data.typing.Sample:
        image = torch.rand([1, 128, 128])
        image = tv_tensors.Image(image)

        return dict(
            image=image, name=sample[0], target=self.target(sample), metadata=sample[2]
        )

    def target(self, sample: typing.Any) -> torch.Tensor:
        return torch.FloatTensor([sample[1]])


def id_function(val):
    if isinstance(val, dict):
        return str(val)
    return repr(val)


@pytest.mark.parametrize(
    "parallel,multiprocessing_context",
    [
        (-1, None),
        (1, None),
        (2, None),
        (4, None),
        (1, "spawn"),
        (2, "spawn"),
        (4, "spawn"),
    ],
    ids=id_function,  # just changes how pytest prints it
)
def test_cached_dataset(parallel, multiprocessing_context):
    from mednet.data.datamodule import CachedDataset

    dataset = CachedDataset(
        raw_dataset=_raw_dataset,
        loader=_RawDataLoader(),
        parallel=parallel,
        multiprocessing_context=multiprocessing_context,
        disable_pbar=True,
    )

    # tests targets
    assert len(dataset.targets()) == _NUM_SAMPLES

    # checks __len__
    assert len(dataset) == _NUM_SAMPLES

    # checks __iter__ works
    # and returns in due order
    for loaded_sample, raw_sample in zip(dataset, _raw_dataset):
        assert loaded_sample["name"] == raw_sample[0]
        assert loaded_sample["target"].item() == raw_sample[1]
        assert loaded_sample["metadata"] == raw_sample[2]

    # checks __getitem__
    for k, raw_sample in enumerate(_raw_dataset):
        loaded_sample = dataset[k]
        assert loaded_sample["name"] == raw_sample[0]
        assert loaded_sample["target"].item() == raw_sample[1]
        assert loaded_sample["metadata"] == raw_sample[2]



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""REFUGE for optic disc and cup segmentation.

The dataset consists of 1200 color fundus photographs, created for a MICCAI
challenge. The goal of the challenge is to evaluate and compare automated
algorithms for glaucoma detection and optic disc/cup segmentation on a common
dataset of retinal fundus images.

* Database reference (including train/dev/test split): :cite:p:`noauthor_refuge_nodate`

.. warning::

   The original directory ``Training400/AMD`` in REFUGE is considered to be
   replaced by an updated version provided by the `AMD Grand-Challenge`_ (with
   matching names).

   The changes concerns images ``A0012.jpg``, which was corrupted in REFUGE,
   and ``A0013.jpg``, which only exists in the AMD Grand-Challenge version.

Data specifications:

* Raw data input (on disk):

  * RGB images encoded in JPG format with varying resolution.  Training images
    are (HxW) 2056 x 2124 pixels; Validation (and test) images are 1634 x 1634
    pixels.
  * Vessel annotations are encoded as BMP images with the same resolution as
    input samples.
  * Masks for the eye fundus are provided by this package.
  * Total samples: 1200 distributed as 400 (training), 400 (validation) and 400
    (test).

* Output sample:

    * Image: Load raw TIFF images with :py:mod:`PIL`, with auto-conversion to RGB.
    * Vessel annotations: Load annotations with :py:mod:`PIL`, with
      auto-conversion to mode ``1`` with no dithering.
    * Eye fundus mask: Load mask with :py:mod:`PIL`, with
      auto-conversion to mode ``1`` with no dithering.

Splits ``optic-disc`` and ``cup`` contain annotations for optic-disc or cup
segmentation.

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import importlib.resources
import importlib.resources.abc
import os
import pathlib
import typing

import PIL.Image
import torch
from torchvision import tv_tensors
from torchvision.transforms.v2.functional import to_dtype, to_image

from ...models.transforms import crop_image_to_mask
from ...utils.rc import load_rc
from ..datamodule import CachingDataModule
from ..split import JSONDatabaseSplit
from ..typing import RawDataLoader as BaseDataLoader
from ..typing import Sample

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name to refer to this database."""

CONFIGURATION_KEY_DATADIR = "datadir." + DATABASE_SLUG
"""Key to search for in the configuration file for the root directory of this
database."""


class RawDataLoader(BaseDataLoader):
    """A specialized raw-data-loader for the drishtigs1 dataset.

    Parameters
    ----------
    target_type
        Indicate whether to use the "cup" or "disc" target.
    """

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self, target_type: str):
        self.datadir = pathlib.Path(
            load_rc().get(CONFIGURATION_KEY_DATADIR, os.path.realpath(os.curdir))
        )
        self.target_type = target_type

    def sample(self, sample: typing.Any) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing path suffixes to the sample image, target, and mask
            to be loaded, within the dataset root folder.

        Returns
        -------
            The sample representation.
        """

        image = PIL.Image.open(self.datadir / sample[0]).convert(mode="RGB")
        image = to_dtype(to_image(image), torch.float32, scale=True)

        target = self.target(sample)

        assert sample[2] is not None
        mask_path = (
            importlib.resources.files(__package__) / "masks" / DATABASE_SLUG / sample[2]
        )
        with importlib.resources.as_file(mask_path) as path:
            mask = PIL.Image.open(path).convert(mode="1", dither=None)
            mask = to_dtype(to_image(mask), torch.float32, scale=True)

        image = tv_tensors.Image(crop_image_to_mask(image, mask))
        target = tv_tensors.Mask(crop_image_to_mask(target, mask))
        mask = tv_tensors.Mask(crop_image_to_mask(mask, mask))

        return dict(image=image, target=target, mask=mask, name=sample[0])

    def target(self, sample: typing.Any) -> torch.Tensor:
        """Load only sample target from its raw representation.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The label corresponding to the specified sample, encapsulated as a
            torch float tensor.
        """

        if self.target_type == "disc":
            target = (
                PIL.Image.open(self.datadir / sample[1])
                .convert(mode="RGB", dither=None)
                .convert("L")
                .point(lambda p: p <= 150, mode="1")
            )

        elif self.target_type == "cup":
            target = (
                PIL.Image.open(self.datadir / sample[1])
                .convert(mode="RGB", dither=None)
                .convert("L")
                .point(lambda p: p <= 100, mode="1")
            )

        else:
            raise ValueError(
                f"Target type {self.target_type} is not an option. "
                f"Available options are 'cup' and 'disc'."
            )

        return to_dtype(to_image(target), torch.float32, scale=True)


class DataModule(CachingDataModule):
    """REFUGE for optic disc and cup segmentation.

    Parameters
    ----------
    split_path
        Path or traversable (resource) with the JSON split description to load.
    target_type
        Indicate whether to use the "cup" or "disc" target.
    """

    def __init__(
        self,
        split_path: pathlib.Path | importlib.resources.abc.Traversable,
        target_type: str,
    ):
        super().__init__(
            database_split=JSONDatabaseSplit(split_path),
            raw_data_loader=RawDataLoader(target_type),
            database_name=DATABASE_SLUG,
            split_name=split_path.name.rsplit(".", 2)[0],
            task="segmentation",
            num_classes=1,
        )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib

import clapper.click
import clapper.logging
import click

from ..click import ConfigCommand

logger = clapper.logging.setup(
    __name__.split(".")[0], format="%(levelname)s: %(message)s"
)


@click.command(
    entry_point_group="mednet.config",
    cls=ConfigCommand,
    epilog="""Examples:

1. Dump annotations for a dataset after pre-processing on a particular directory:

   .. code:: sh

      mednet segment dump-annotations -vv lwnet drive-2nd --output-folder=path/to/annotations

""",
)
@click.option(
    "--output-folder",
    "-o",
    help="Directory in which to save predictions (created if does not exist)",
    required=True,
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        writable=True,
        path_type=pathlib.Path,
    ),
    default="predictions",
    show_default=True,
    cls=clapper.click.ResourceOption,
)
@click.option(
    "--model",
    "-m",
    help="""A lightning module instance that will be used to retrieve
    pre-processing transforms.""",
    required=True,
    cls=clapper.click.ResourceOption,
)
@click.option(
    "--datamodule",
    "-d",
    help="""A lightning DataModule that will be asked for prediction data
    loaders. Typically, this includes all configured splits in a DataModule,
    however this is not a requirement.  A DataModule that returns a single
    dataloader for prediction (wrapped in a dictionary) is acceptable.""",
    required=True,
    cls=clapper.click.ResourceOption,
)
@clapper.click.verbosity_option(logger=logger, expose_value=False)
def dump_annotations(
    output_folder, model, datamodule, **_
) -> None:  # numpydoc ignore=PR01
    """Dump annotations in a given folder, after pre-processing."""

    from ...engine.segment.dumper import run
    from ...engine.trainer import validate_model_datamodule
    from ..utils import save_json_metadata, save_json_with_backup

    validate_model_datamodule(model, datamodule)
    assert datamodule.task == "segmentation"

    # sets-up the data module
    datamodule.model_transforms = list(model.model_transforms)
    datamodule.batch_size = 1
    datamodule.parallel = -1
    datamodule.prepare_data()
    datamodule.setup(stage="predict")

    # stores all information we can think of, to reproduce this later
    save_json_metadata(
        output_file=output_folder / "annotations.meta.json",
        output_folder=output_folder,
        model=model,
        datamodule=datamodule,
    )

    json_data = run(datamodule, output_folder)

    base_file = output_folder / "annotations.json"
    save_json_with_backup(base_file, json_data)
    logger.info(f"Annotations saved to `{str(base_file)}`")



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Faster R-CNN object detection (and classification) network architecture, from :cite:p:`ren_faster_2017`."""

import logging
import typing

import torch
import torch.nn
import torch.optim
import torch.optim.lr_scheduler
import torch.optim.optimizer
import torch.utils.data
import torchvision.models as models

from ...data.typing import TransformSequence
from ..typing import Checkpoint
from .model import Model

logger = logging.getLogger(__name__)


class FasterRCNN(Model):
    """Faster R-CNN object detection (and classification) network architecture, from :cite:p:`ren_faster_2017`.

    Parameters
    ----------
    optimizer_type
        The type of optimizer to use for training.
    optimizer_arguments
        Arguments to the optimizer after ``params``.
    scheduler_type
        The type of scheduler to use for training.
    scheduler_arguments
        Arguments to the scheduler after ``params``.
    model_transforms
        An optional sequence of torch modules containing transforms to be applied on the
        input **before** it is fed into the network.
    augmentation_transforms
        An optional sequence of torch modules containing transforms to be applied on the
        input **before** it is fed into the network.
    pretrained
        If set to True, loads pretrained model weights during initialization, else
        trains a new model.
    num_classes
        Number of outputs (classes) for this model. Do not account for the background
        (we compensate internally).
    variant
        One of the torchvision supported variants.
    """

    def __init__(
        self,
        optimizer_type: type[torch.optim.Optimizer] = torch.optim.SGD,
        optimizer_arguments: dict[str, typing.Any] = dict(
            lr=0.005, momentum=0.9, weight_decay=0.0005
        ),
        scheduler_type: type[torch.optim.lr_scheduler.LRScheduler]
        | None = torch.optim.lr_scheduler.StepLR,
        scheduler_arguments: dict[str, typing.Any] = dict(step_size=3, gamma=0.1),
        model_transforms: TransformSequence | None = None,
        augmentation_transforms: TransformSequence | None = None,
        pretrained: bool = False,
        num_classes: int = 1,
        variant: typing.Literal[
            "resnet50-v1", "resnet50-v2", "mobilenetv3-large", "mobilenetv3-small"
        ] = "mobilenetv3-small",
    ):
        super().__init__(
            name=f"faster-rcnn[{variant}]",
            loss_type=None,
            loss_arguments=None,
            optimizer_type=optimizer_type,
            optimizer_arguments=optimizer_arguments,
            scheduler_type=scheduler_type,
            scheduler_arguments=scheduler_arguments,
            model_transforms=model_transforms,
            augmentation_transforms=augmentation_transforms,
            num_classes=num_classes,
        )

        self.pretrained = pretrained
        self.variant = variant

        # Load pretrained model
        weights = None
        if pretrained:
            from ..normalizer import make_cocov1_normalizer

            Model.normalizer.fset(self, make_cocov1_normalizer())  # type: ignore[attr-defined]

            logger.info(f"Loading pretrained `{self.name}` model weights")
            match self.variant:
                case "resnet50-v1":
                    weights = "FasterRCNN_ResNet50_FPN_Weights.COCO_V1"
                case "resnet50-v2":
                    weights = "FasterRCNN_ResNet50_FPN_V2_Weights.COCO_V1"
                case "mobilenetv3-large":
                    weights = "FasterRCNN_MobileNet_V3_Large_FPN_Weights.COCO_V1"
                case "mobilenetv3-small":
                    weights = "FasterRCNN_MobileNet_V3_Large_320_FPN_Weights.COCO_V1"

        match self.variant:
            case "resnet50-v1":
                self.model = models.detection.fasterrcnn_resnet50_fpn(weights=weights)
            case "resnet50-v2":
                self.model = models.detection.fasterrcnn_resnet50_fpn_v2(
                    weights=weights
                )
            case "mobilenetv3-large":
                self.model = models.detection.fasterrcnn_mobilenet_v3_large_fpn(
                    weights=weights
                )
            case "mobilenetv3-small":
                self.model = models.detection.fasterrcnn_mobilenet_v3_large_320_fpn(
                    weights=weights
                )

        # Instantiates model and adapts output features
        self.num_classes = num_classes

    @Model.num_classes.setter  # type: ignore[attr-defined]
    def num_classes(self, v: int) -> None:
        # Faster R-CNN models will have num_classes + 1 outputs accounting for the
        # background
        v += 1

        if self.model.roi_heads.box_predictor.cls_score.out_features != v:
            if self.pretrained:
                logger.info(
                    f"Resetting `{self.name}` pretrained classifier "
                    f"layer weights due to change in output size "
                    f"({self.model.roi_heads.box_predictor.cls_score.out_features} -> {v})"
                )
            self.model.roi_heads.box_predictor = (
                models.detection.faster_rcnn.FastRCNNPredictor(
                    self.model.roi_heads.box_predictor.cls_score.in_features, v
                )
            )
        self._num_classes = v

    def on_save_checkpoint(self, checkpoint: Checkpoint) -> None:
        checkpoint["variant"] = self.variant
        super().on_save_checkpoint(checkpoint)

    def on_load_checkpoint(self, checkpoint: Checkpoint) -> None:
        # setup model type
        self.variant = checkpoint.get("variant", "mobilenetv3-small")
        match self.variant:
            case "resnet50-v1":
                self.model = models.detection.fasterrcnn_resnet50_fpn()
            case "resnet50-v2":
                self.model = models.detection.fasterrcnn_resnet50_fpn_v2()
            case "mobilenetv3-large":
                self.model = models.detection.fasterrcnn_mobilenet_v3_large_fpn()
            case "mobilenetv3-small":
                self.model = models.detection.fasterrcnn_mobilenet_v3_large_320_fpn()

        # reset number of classes if need be
        num_classes = checkpoint["state_dict"][
            "model.roi_heads.box_predictor.cls_score.bias"
        ].shape[0]

        if num_classes != self.num_classes:
            logger.debug(
                f"Resetting number-of-output-classes at `{self.name}` model from "
                f"{self.num_classes} to {num_classes} while loading checkpoint."
            )
        self.num_classes = num_classes - 1

        super().on_load_checkpoint(checkpoint)

    def forward(
        self,
        images: list[torch.Tensor],
        targets: list[dict[str, torch.Tensor]] | None = None,
    ):
        """Forward the input tensor through the network, producing a prediction.

        Parameters
        ----------
        images
            Input images, to be analyzed or trained on.
        targets
            Targets for the current input images.  Targets should be passed only if
            *training*. It should be alist of dictionaries, each corresponding to one of
            the input images in ``images``.  Each dictionary should have two keys:
            ``boxes``, that contains the bounding boxes associated with the image, and
            ``labels``, that contains the labels associated with each bounding box.

        Returns
        -------
            A list with various dictionaries, each referring to one of the
        """
        images = [self.normalizer(k) for k in images]

        if targets is None:
            return self.model(images)

        return self.model(images, targets)

    def training_step(self, batch, batch_idx):
        del batch_idx

        # debug code to inspect images by eye:
        # from torchvision.transforms.v2.functional import to_pil_image
        # for k in batch["image"]:
        #    to_pil_image(k).show()
        #    __import__("pdb").set_trace()

        # during training, detection models __call__() take 2 lists as inputs:
        # * list of images
        # * list of dictionaries with "boxes" and "labels"
        #
        # -> model returns a Dict[Tensor] during training, containing the
        #    classification and regression losses for both the RPN and the R-CNN.
        result = self.forward(
            [k for k in self._augmentation_transforms(batch["image"])],
            [
                {"boxes": k1, "labels": k2.long()}
                for (k1, k2) in zip(
                    self._augmentation_transforms(batch["target"]), batch["labels"]
                )
            ],
        )
        return sum(result.values())

    def validation_step(self, batch, batch_idx, dataloader_idx=0):
        del batch_idx, dataloader_idx  # satisfies linter

        # need to be in "train" mode to calculate losses
        self.model.train()
        with torch.no_grad():
            result = self(
                [k for k in batch["image"]],
                [
                    {"boxes": k1, "labels": k2.long()}
                    for (k1, k2) in zip(batch["target"], batch["labels"])
                ],
            )
        return sum(result.values())

    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        del batch_idx, dataloader_idx  # satisfies linter

        # debug code to inspect images by eye:
        # from torchvision.transforms.v2.functional import to_pil_image
        # for k in batch["image"]:
        #    to_pil_image(k).show()
        #    __import__("pdb").set_trace()

        # during inference, detection models __call__() take 1 list as input:
        # * list of images
        #
        # -> model returns the post-processed predictions as a
        #    list[dict[torch.Tensor]], one for each input image
        return self([k for k in batch["image"]])



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib
import typing

import click
from clapper.click import ResourceOption, verbosity_option

from ....models.classify.typing import SaliencyMapAlgorithm
from ...click import ConfigCommand
from ...logging import setup_cli_logger

logger = setup_cli_logger()


@click.command(
    entry_point_group="mednet.config",
    cls=ConfigCommand,
    epilog="""Examples:

1. Calculate the ROAD scores using GradCAM, for an existing dataset
   configuration and stores them in .json files:

   .. code:: sh

      mednet classify saliency completeness -vv pasa tbx11k-v1-healthy-vs-atb --saliency-map-algorithm="gradcam" --device="cuda" --weight=path/to/model-at-lowest-validation-loss.ckpt --output-json=path/to/completeness-scores.json

""",
)
@click.option(
    "--model",
    "-m",
    help="""A lightning module instance implementing the network architecture
    (not the weights, necessarily) to be used for inference.  Currently, only
    supports pasa and densenet models.""",
    required=True,
    type=click.UNPROCESSED,
    cls=ResourceOption,
)
@click.option(
    "--datamodule",
    "-d",
    help="""A lightning DataModule that will be asked for prediction DataLoaders.
    Typically, this includes all configured splits in a DataModule,
    however this is not a requirement.  A DataModule that returns a single
    DataLoader for prediction (wrapped in a dictionary) is acceptable.""",
    required=True,
    type=click.UNPROCESSED,
    cls=ResourceOption,
)
@click.option(
    "--output-folder",
    "-o",
    help="""Name of the folder where to store the output .json file containing all
    measures.""",
    required=True,
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        path_type=pathlib.Path,
        writable=True,
    ),
    default="results",
    show_default=True,
    cls=ResourceOption,
)
@click.option(
    "--device",
    "-x",
    help='A string indicating the device to use (e.g. "cpu" or "cuda:0")',
    show_default=True,
    required=True,
    default="cpu",
    cls=ResourceOption,
)
@click.option(
    "--cache-samples/--no-cache-samples",
    help="If set to True, loads the sample into memory, "
    "otherwise loads them at runtime.",
    required=True,
    show_default=True,
    default=False,
    cls=ResourceOption,
)
@click.option(
    "--weight",
    "-w",
    help="""Path or URL to pretrained model file (`.ckpt` extension),
    corresponding to the architecture set with `--model`.  Optionally, you may
    also pass a directory containing the result of a training session, in which
    case either the best (lowest validation) or latest model will be loaded.""",
    required=True,
    cls=ResourceOption,
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=True,
        readable=True,
        path_type=pathlib.Path,
    ),
)
@click.option(
    "--parallel",
    "-P",
    help="""Use multiprocessing for data loading processing: if set to -1
    (default), disables multiprocessing.  Set to 0 to enable as many data
    processing instances as processing cores available in the system.  Set to
    >= 1 to enable that many multiprocessing instances.  Note that if you
    activate this option, then you must use --device=cpu, as using a GPU
    concurrently is not supported.""",
    type=click.IntRange(min=-1),
    show_default=True,
    required=True,
    default=-1,
    cls=ResourceOption,
)
@click.option(
    "--saliency-map-algorithm",
    "-s",
    help="""Saliency map algorithm to be used.""",
    type=click.Choice(
        typing.get_args(SaliencyMapAlgorithm),
        case_sensitive=False,
    ),
    default="gradcam",
    show_default=True,
    cls=ResourceOption,
)
@click.option(
    "--target-class",
    "-C",
    help="""This option should only be used with multiclass models.  It
    defines the class to target for saliency estimation. Can be either set to
    "all" or "highest". "highest" (the default), means only saliency maps for
    the class with the highest activation will be generated.""",
    required=False,
    type=click.Choice(
        ["highest", "all"],
        case_sensitive=False,
    ),
    default="highest",
    cls=ResourceOption,
)
@click.option(
    "--positive-only/--no-positive-only",
    "-z/-Z",
    help="""If set, and the model chosen has a single output (binary), then
    saliency maps will only be generated for samples of the positive class.
    This option has no effect for multiclass models.""",
    default=False,
    cls=ResourceOption,
)
@click.option(
    "--percentile",
    "-e",
    help="""One or more percentiles (percent x100) integer values indicating
    the proportion of pixels to perturb in the original image to calculate both
    MoRF and LeRF scores.""",
    multiple=True,
    default=[20, 40, 60, 80],
    show_default=True,
    cls=ResourceOption,
)
@click.option(
    "--only-dataset",
    "-S",
    help="""If set, will only run the command for the named dataset on the
    provided datamodule, skipping any other dataset.""",
    cls=ResourceOption,
)
@click.option(
    "--plot/--no-plot",
    "-P",
    help="""If set, then also produces figures containing the plots of
    score histograms.""",
    required=True,
    show_default=True,
    default=True,
    cls=ResourceOption,
)
@verbosity_option(logger=logger, expose_value=False)
def completeness(
    model,
    datamodule,
    output_folder,
    device,
    cache_samples,
    weight,
    parallel,
    saliency_map_algorithm,
    target_class,
    positive_only,
    percentile,
    only_dataset,
    plot,
    **_,
) -> None:  # numpydoc ignore=PR01
    """Evaluate saliency map algorithm completeness using RemOve And Debias
    (ROAD).

    For the selected saliency map algorithm, evaluates the completeness of
    explanations using the RemOve And Debias (ROAD) algorithm. The ROAD
    algorithm was first described at :cite:p:`rong_consistent_2022`. It estimates explainability
    (in the completeness sense) of saliency mapping algorithms by substituting
    relevant pixels in the input image by a local average, re-running
    prediction on the altered image, and measuring changes in the output
    classification score when said perturbations are in place.  By substituting
    most or least relevant pixels with surrounding averages, the ROAD algorithm
    estimates the importance of such elements in the produced saliency map.  As
    2023, this measurement technique is considered to be one of the
    state-of-the-art metrics of explainability.

    This program outputs a .json file containing the ROAD evaluations (using
    most-relevant-first, or MoRF, and least-relevant-first, or LeRF for each
    sample in the DataModule. Values for MoRF and LeRF represent averages by
    removing 20, 40, 60 and 80% of most or least relevant pixels respectively
    from the image, and averaging results for all these percentiles.

    .. note::

       This application is relatively slow when processing a large DataModule
       with many (positive) samples.
    """

    import matplotlib.backends.backend_pdf

    from ....engine.classify.saliency.completeness import run
    from ....engine.classify.saliency.utils import make_plots, make_table
    from ....engine.device import DeviceManager
    from ....engine.trainer import validate_model_datamodule
    from ....models.classify.model import Model as ClassificationModel
    from ....utils.checkpointer import get_checkpoint_to_run_inference
    from ...utils import save_json_metadata, save_json_with_backup

    if device in ("cuda", "mps") and (parallel == 0 or parallel > 1):
        raise RuntimeError(
            f"The number of multiprocessing instances is set to {parallel} and "
            f"you asked to use a GPU (device = `{device}`). The currently "
            f"implementation can only handle a single GPU.  Either disable GPU "
            f"utilisation or set the number of multiprocessing instances to "
            f"one, or disable multiprocessing entirely (ie. set it to -1).",
        )

    validate_model_datamodule(model, datamodule)
    assert isinstance(model, ClassificationModel)

    device_manager = DeviceManager(device)

    datamodule.cache_samples = cache_samples
    datamodule.model_transforms = list(model.model_transforms)
    datamodule.batch_size = 1
    datamodule.parallel = parallel
    datamodule.prepare_data()
    datamodule.setup(stage="predict")

    output_json = output_folder / "completeness.json"

    if weight.is_dir():
        weight = get_checkpoint_to_run_inference(weight)

    logger.info(f"Loading checkpoint from `{weight}`...")
    model = type(model).load_from_checkpoint(weight, strict=False)

    # stores all information we can think of, to reproduce this later
    save_json_metadata(
        output_file=output_json.with_suffix(".meta.json"),
        model=model,
        datamodule=datamodule,
        output_json=output_json,
        device=device,
        cache_samples=cache_samples,
        weight=weight,
        parallel=parallel,
        saliency_map_algorithm=saliency_map_algorithm,
        target_class=target_class,
        positive_only=positive_only,
        percentile=percentile,
        only_dataset=only_dataset,
        plot=plot,
    )

    logger.info(
        f"Evaluating RemOve And Debias (ROAD) average scores for "
        f"algorithm `{saliency_map_algorithm}` with percentiles "
        f"`{', '.join([str(k) for k in percentile])}`...",
    )
    results = run(
        model=model,
        datamodule=datamodule,
        device_manager=device_manager,
        saliency_map_algorithm=saliency_map_algorithm,
        target_class=target_class,
        positive_only=positive_only,
        percentiles=percentile,
        parallel=parallel,
        only_dataset=only_dataset,
    )

    logger.info(f"Saving output file to `{str(output_json)}`...")
    save_json_with_backup(output_json, results)

    table = make_table(
        results=results,
        indexes={3: "ROAD (MoRF)", 4: "ROAD (LeRF)", 5: "ROAD (avg)"},
        format_="rst",
    )
    output_table = output_json.with_suffix(".rst")
    logger.info(f"Saving output summary table to `{str(output_table)}`...")
    with output_table.open("w") as f:
        f.write(table)
    click.echo(table)

    # Plots histograms, if the user asked to do so.
    if plot:
        figure_path = output_json.with_suffix(".pdf")
        logger.info(f"Saving plots to `{str(figure_path)}`...")
        with matplotlib.backends.backend_pdf.PdfPages(figure_path) as pdf:
            for fig in make_plots(
                results=results,
                indexes={3: "ROAD (MoRF)", 4: "ROAD (LeRF)", 5: "ROAD (avg)"},
            ):
                pdf.savefig(fig)



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Montgomery database for lung segmentation.

The standard digital image database for Tuberculosis was created by the National
Library of Medicine, Maryland, USA in collaboration with Shenzhen No.3 People’s
Hospital, Guangdong Medical College, Shenzhen, China. The Chest X-rays are from

* Database reference: :cite:p:`jaeger_two_2014`, :cite:p:`gaal_attention_2020`
* Original resolution (height x width or width x height): 4020x4892 px or
  4892x4020 px

Data specifications:

* Raw data input (on disk):

  * PNG images 8 bit grayscale issued from digital radiography machines
  * Original resolution (height x width or width x height): 4020x4892 px or
    4892x4020 px
  * Samples: 138 images and associated labels

* Output image:

  * Transforms:

    * Load raw PNG with :py:mod:`PIL`
    * Convert to torch tensor

  * Final specifications

    * image: Grayscale, encoded as a single plane tensor, 32-bit floats,
      original size.
    * target: A binary mask containing ones where lungs are in the original
      image, otherwise, zeroes.
    * mask: Binary, with all ones (no specific mask)

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import importlib.resources.abc
import os
import pathlib
import typing

import numpy as np
import PIL.Image
import torch
from torchvision import tv_tensors
from torchvision.transforms.v2.functional import to_dtype, to_image

from ...utils.rc import load_rc
from ..datamodule import CachingDataModule
from ..split import JSONDatabaseSplit
from ..typing import RawDataLoader as BaseDataLoader
from ..typing import Sample

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name to refer to this database."""

CONFIGURATION_KEY_DATADIR = "datadir." + DATABASE_SLUG
"""Key to search for in the configuration file for the root directory of this
database."""


class RawDataLoader(BaseDataLoader):
    """A specialized raw-data-loader for the Montgomery dataset."""

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self):
        self.datadir = pathlib.Path(
            load_rc().get(CONFIGURATION_KEY_DATADIR, os.path.realpath(os.curdir))
        )

    def sample(self, sample: typing.Any) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing the
            sample label.

        Returns
        -------
            The sample representation.
        """

        image = PIL.Image.open(self.datadir / sample[0]).convert(mode="RGB")
        image = to_dtype(to_image(image), torch.float32, scale=True)

        target = self.target(sample)

        # use image as a base since target() can be overriden by child class
        mask = torch.ones((1, image.shape[-2], image.shape[-1]), dtype=torch.float32)

        image = tv_tensors.Image(image)
        target = tv_tensors.Mask(target)
        mask = tv_tensors.Mask(mask)

        return dict(image=image, target=target, mask=mask, name=sample[0])

    def target(self, sample: typing.Any) -> torch.Tensor:
        """Load only sample target from its raw representation.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The label corresponding to the specified sample, encapsulated as a
            torch float tensor.
        """

        # Combines left and right lung masks into a single tensor
        left = PIL.Image.open(self.datadir / sample[1]).convert(mode="1", dither=None)
        right = PIL.Image.open(self.datadir / sample[2]).convert(mode="1", dither=None)
        target = np.ma.mask_or(np.asarray(left), np.asarray(right))

        return to_dtype(to_image(target), torch.float32, scale=True)


class DataModule(CachingDataModule):
    """Montgomery database for lung segmentation.

    Parameters
    ----------
    split_path
        Path or traversable (resource) with the JSON split description to load.
    """

    def __init__(self, split_path: pathlib.Path | importlib.resources.abc.Traversable):
        super().__init__(
            database_split=JSONDatabaseSplit(split_path),
            raw_data_loader=RawDataLoader(),
            database_name=DATABASE_SLUG,
            split_name=split_path.name.rsplit(".", 2)[0],
            task="segmentation",
            num_classes=1,
        )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Prediction engine for object detection tasks."""

import logging
import typing

import lightning.pytorch.callbacks
import torch.utils.data

from ...engine.device import DeviceManager
from ...models.detect.typing import Prediction, PredictionSplit
from ...utils.string import rewrap

logger = logging.getLogger(__name__)


class _JSONMetadataCollector(lightning.pytorch.callbacks.BasePredictionWriter):
    """Collects further sample metadata to store with predictions.

    This object collects further sample metadata we typically keep with
    predictions.

    Parameters
    ----------
    write_interval
        When will this callback be active.
    """

    def __init__(
        self,
        write_interval: typing.Literal["batch", "epoch", "batch_and_epoch"] = "batch",
    ):
        super().__init__(write_interval=write_interval)
        self._data: list = []

    def write_on_batch_end(
        self,
        trainer: lightning.pytorch.Trainer,
        pl_module: lightning.pytorch.LightningModule,
        prediction: typing.Any,
        batch_indices: typing.Sequence[int] | None,
        batch: typing.Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        """Write batch predictions to disk.

        Parameters
        ----------
        trainer
            The trainer being used.
        pl_module
            The pytorch module.
        prediction
            The actual predictions to record.
        batch_indices
            The relative position of samples on the epoch.
        batch
            The current batch.
        batch_idx
            Index of the batch overall.
        dataloader_idx
            Index of the dataloader overall.
        """
        del trainer, pl_module, batch_indices, batch_idx, dataloader_idx

        for k, sample_pred in enumerate(prediction):
            sample_name: str = batch["name"][k]

            targets = []
            for box, label in zip(batch["target"][k], batch["labels"][k]):
                targets.append(
                    [
                        box.cpu().int().numpy().tolist(),
                        label.cpu().int().numpy().tolist(),
                    ]
                )

            predictions = []
            for box, label, score in zip(
                sample_pred["boxes"], sample_pred["labels"], sample_pred["scores"]
            ):
                predictions.append(
                    [
                        box.cpu().numpy().tolist(),
                        label.cpu().numpy().tolist(),
                        score.cpu().numpy().tolist(),
                    ]
                )

            self._data.append((sample_name, targets, predictions))

    def reset(self) -> list[Prediction]:
        """Summary of written objects.

        Also resets the internal state.

        Returns
        -------
            A list containing a summary of all samples written.
        """
        retval = self._data
        self._data = []
        return retval


def run(
    model: lightning.pytorch.LightningModule,
    datamodule: lightning.pytorch.LightningDataModule,
    device_manager: DeviceManager,
) -> list[Prediction] | list[list[Prediction]] | PredictionSplit | None:
    """Run inference on input data, output predictions.

    Parameters
    ----------
    model
        Neural network model (e.g. faster-rcnn).
    datamodule
        The lightning DataModule to run predictions on.
    device_manager
        An internal device representation, to be used for training and
        validation.  This representation can be converted into a pytorch device
        or a lightning accelerator setup.

    Returns
    -------
        Depending on the return type of the DataModule's
        ``predict_dataloader()`` method:

            * if :py:class:`torch.utils.data.DataLoader`, then returns a
              :py:class:`list` of predictions.
            * if :py:class:`list` of :py:class:`torch.utils.data.DataLoader`, then
              returns a list of lists of predictions, each list corresponding to
              the iteration over one of the dataloaders.
            * if :py:class:`dict` of :py:class:`str` to
              :py:class:`torch.utils.data.DataLoader`, then returns a dictionary
              mapping names to lists of predictions.
            * if ``None``, then returns ``None``.

    Raises
    ------
    TypeError
        If the DataModule's ``predict_dataloader()`` method does not return any
        of the types described above.
    """

    from lightning.pytorch.loggers.logger import DummyLogger

    collector = _JSONMetadataCollector()

    accelerator, devices = device_manager.lightning_accelerator()
    trainer = lightning.pytorch.Trainer(
        accelerator=accelerator,
        devices=devices,
        logger=DummyLogger(),
        callbacks=[collector],
    )

    dataloaders = datamodule.predict_dataloader()

    if isinstance(dataloaders, torch.utils.data.DataLoader):
        logger.info("Running prediction on a single dataloader...")
        trainer.predict(model, dataloaders, return_predictions=False)
        return collector.reset()

    if isinstance(dataloaders, list):
        retval_list = []
        for k, dataloader in enumerate(dataloaders):
            logger.info(f"Running prediction on split `{k}`...")
            trainer.predict(model, dataloader, return_predictions=False)
            retval_list.append(collector.reset())
        return retval_list  # type: ignore

    if isinstance(dataloaders, dict):
        retval_dict = {}
        for name, dataloader in dataloaders.items():
            logger.info(f"Running prediction on `{name}` split...")
            trainer.predict(model, dataloader, return_predictions=False)
            retval_dict[name] = collector.reset()
        return retval_dict  # type: ignore

    if dataloaders is None:
        logger.warning("Datamodule did not return any prediction dataloaders!")
        return None

    # if you get to this point, then the user is returning something that is
    # not supported - complain!
    raise TypeError(
        rewrap(
            f"""Datamodule returned strangely typed prediction dataloaders:
            `{type(dataloaders)}` - if this is not an error, write code to support this
            use-case."""
        )
    )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Functions to upload models and measures to GitLab's experiment registry."""

import configparser
import json
import logging
import os
import pathlib
import re
import shutil
import tempfile
import typing

import gitlab
import mlflow

from ..utils.checkpointer import get_checkpoint_to_run_inference

logger = logging.getLogger(__name__)


def _gitlab_instance_and_token() -> tuple[gitlab.Gitlab, str]:
    """Return an instance of the Gitlab object for remote operations, and the
    user token.

    Returns
    -------
        Gitlab main object and user token
    """

    cfg = pathlib.Path("~/.python-gitlab.cfg").expanduser()
    if cfg.exists():
        gl = gitlab.Gitlab.from_config("idiap", [str(cfg)])
        config = configparser.ConfigParser()
        config.read(cfg)
        token = config["idiap"]["private_token"]

    else:  # ask the user for a token or use one from the current runner
        server = "https://gitlab.idiap.ch"
        token = input(f"{server} (user or project) token: ")
        gl = gitlab.Gitlab(server, private_token=token, api_version="4")

    # tests authentication with given credential.
    gl.auth()

    return gl, token


def _size_in_mb(path: pathlib.Path) -> float:
    """Return the size in megabytes of a file.

    Parameters
    ----------
    path
        Input path to calculate file size from.

    Returns
    -------
        A floating point number for the size of the object in MB.
    """
    return path.stat().st_size / (1024**2)


def _assemble_artifacts(
    experiment_folder: pathlib.Path, upload_limit_mb: int
) -> tempfile.TemporaryDirectory:
    """Assemble artifacts (files) to upload, sanitize filenames and check overall upload
    size.

    The files that are uploaded are the following:

    * train.meta.json: meta information during training
    * trainlog.pdf: evolution of tracked training variables
    * evaluation.json: evaluation results
    * evaluation.meta.json: meta information during evaluation
    * evaluation.rst: evaluation results in table format
    * evaluation.pdf: evaluation plots
    * model checkpoint (variable name)

    Parameters
    ----------
    experiment_folder
        Directory in which to upload results from.
    upload_limit_mb
        Maximim upload size in MB (set to 0 for no limit).

    Returns
    -------
        Temporary directory where the important files from the experiment
        folder have been assembled and names sanitized.

    Raises
    ------
    AssertionError
        In case one of the necessary files that are typically uploaded is
        missing.
    RuntimeError
        In case the total size of the temporary directory contents is larger,
        in MB, than the ``upload_limit_mb``.
    """

    logger.info(f"Assembling files from {experiment_folder}...")

    # get train files
    train_folder = experiment_folder
    train_log_file = train_folder / "trainlog.pdf"
    train_meta_file = train_folder / "train.meta.json"
    train_model_file = get_checkpoint_to_run_inference(train_folder)
    train_files = [train_meta_file, train_model_file, train_log_file]

    # get evaluation files
    evaluation_file = experiment_folder / "evaluation.json"
    evaluation_meta_file = experiment_folder / "evaluation.meta.json"
    evaluation_meta_file = experiment_folder / "evaluation.rst"
    evaluation_log_file = experiment_folder / "evaluation.pdf"
    evaluation_files = [
        evaluation_file,
        evaluation_meta_file,
        evaluation_log_file,
    ]

    # checks all files exist
    for f in train_files + evaluation_files:
        assert f.exists(), f"Missing file `{f}` - cannot upload artifact"

    # checks for maximum upload limit
    total_size_mb = sum([_size_in_mb(f) for f in train_files + evaluation_files])
    if upload_limit_mb != 0 and total_size_mb > upload_limit_mb:
        raise RuntimeError(
            f"Total size of upload ({total_size_mb:.2f} MB) exceeds "
            f"permitted maximum ({upload_limit_mb:.2f} MB)."
        )

    retval = tempfile.TemporaryDirectory()
    tmpdir_path = pathlib.Path(retval.name)
    for f in train_files + evaluation_files:
        clean = tmpdir_path / f.parts[-1].replace("=", "-")
        shutil.copy2(f, clean)
        logger.debug(f"`{str(f)}` -> `{str(clean)}` ({_size_in_mb(f):.2f} MB)")

    logger.info(f"Total size of files at {retval.name} = {total_size_mb:.2f} MB")
    return retval


def _check_version(version: str) -> None:
    """Check if a provided version number is not dirty compatible.

    Checks if the version of the model to be uploaded is not dirty (i.e. ends
    with something like ``.d20240807``).

    Parameters
    ----------
    version
        A string indicating the model version (e.g. "1.0.0").
    """

    assert not re.search(r"\.d[0-9]*$", version), (
        f"Incompatible model version ({version}) - you should NOT "
        f"upload models from `dirty` repositories"
    )


def _assemble_parameters(basedir: pathlib.Path) -> dict[str, typing.Any]:
    """Assemble parameters to log to experiment.

    Parameters are forcebly converted to string representations via the MLflow
    interface.

    Parameters
    ----------
    basedir
        Base directory where to find the file ``train.meta.json``.  Typically,
        this is the experiment folder.

    Returns
    -------
        A dictionary that maps strings to any value that, itself, can be
        converted to a string.  The MLflow interface will take care of this.
    """

    train_meta_file = basedir / "train.meta.json"
    with train_meta_file.open("r") as meta_file:
        train_data = json.load(meta_file)

    _check_version(train_data["package-version"])

    # get lowest validation epoch
    train_model_file = get_checkpoint_to_run_inference(basedir)
    best_epoch = int(str(train_model_file).split(".")[0].rsplit("=", 1)[1])

    return {
        "package version": train_data["package-version"],
        "batch size": train_data["batch-size"],
        "batch accumulations": train_data["accumulate-grad-batches"],
        "epochs": train_data["epochs"],
        "model epoch": best_epoch,
    }


def _assemble_metrics(basedir: pathlib.Path, names: list[str]) -> dict[str, float]:
    """Assemble metrics to log to experiment.

    Metrics are float values that can use to measure the performance of a
    model.

    Parameters
    ----------
    basedir
        Base directory where to find the file ``evaluation.json``.  Typically,
        this is the experiment folder.
    names
        A list of metrics we are interested in fetching from the evaluation
        file, and export to GitLab.

    Returns
    -------
        A dictionary that maps strings to floating point values.
    """

    evaluation_file = basedir / "evaluation.json"
    with evaluation_file.open("r") as f:
        evaluation_data = json.load(f)

    return {k: v for k, v in evaluation_data["test"].items() if k in names}


def _user_or_default_names(
    basedir: pathlib.Path, experiment_name: str, run_name: str
) -> tuple[str, str]:
    """Assert user-provided experiment and run names or defaults.

    Parameters
    ----------
    basedir
        Base directory where to find the file ``train.meta.json``.  Typically,
        this is the experiment folder.
    experiment_name
        User provided experiment name.  If empty, then a default experiment
        name will be proposed.
    run_name
        User provided run name.  If empty, then a default run name using the
        experiment date will be proposed.

    Returns
    -------
        A tuple containing the experiment and run name to be used.
    """

    train_meta_file = basedir / "train.meta.json"
    with train_meta_file.open("r") as meta_file:
        train_data = json.load(meta_file)

    return (
        experiment_name or f"{train_data['model-name']}-{train_data['database-name']}",
        run_name or train_data["datetime"],
    )


def _upload_ml_experiment(
    project_path: str,
    experiment_name: str,
    run_name: str,
    artifact_path: pathlib.Path,
    parameters: dict[str, typing.Any],
    metrics: dict[str, float],
) -> None:
    """Upload to GitLab using the Machine Learning Experiment Tracking
    interface.

    Information about the ML Experiment Tracking interface can be found at
    https://docs.gitlab.com/ee/user/project/ml/experiment_tracking/

    Parameters
    ----------
    project_path
        Path to the project where to upload model entries.
    experiment_name
        A string indicating the experiment name (e.g. "exp-pasa-mc" or "exp-densenet-mc-ch").
    run_name
        A string indicating the run name (e.g. "run-1").
    artifact_path
        A base directory in which all contained files will be uploaded as
        artifacts to the experiment entry.
    parameters
        All experiment parameters (``str`` -> ``str``) to log to the
        experiment table.
    metrics
        All experiment metrics (``str`` -> ``float``) to log to the experiment
        table.
    """

    logger.info("Retrieving GitLab credentials for access to hosted MLFlow server...")
    gitlab, token = _gitlab_instance_and_token()
    project = gitlab.projects.get(project_path)
    os.environ["MLFLOW_TRACKING_TOKEN"] = token
    os.environ["MLFLOW_TRACKING_URI"] = (
        gitlab.api_url + f"/projects/{project.id}/ml/mlflow"
    )

    logger.info(
        f"Uploading entry `{run_name}` to experiment `{experiment_name}` "
        f"on GitLab project `{project.name_with_namespace}` (id: {project.id})..."
    )
    exp_meta = mlflow.set_experiment(experiment_name=experiment_name)

    with mlflow.start_run(run_name=run_name):
        logger.info("Uploading parameters...")
        for key, value in parameters.items():
            logger.info(f'[parameter] "{key}" = "{str(value)}"')
            mlflow.log_param(key, value, synchronous=True)

        logger.info("Uploading metrics...")
        for key, value in metrics.items():
            logger.info(f'[metric] "{key}" = {value:.3g}')
            mlflow.log_metric(key, value, synchronous=True)

        logger.info("Uploading artifacts (files)...")
        for f in artifact_path.glob("*.*"):
            logger.info(f'[artifact] "{str(f)}" ({_size_in_mb(f):.2f} MB)')
            mlflow.log_artifact(str(f))

    logger.info(
        f"Visit {gitlab.url}/{project.path_with_namespace}/-/ml/experiments/{exp_meta.experiment_id}"
    )


def run(
    project_path: str,
    experiment_folder: pathlib.Path,
    experiment_name: str,
    run_name: str,
    metrics: list[str],
    upload_limit_mb: int,
) -> None:
    """Upload results from an experiment folder to GitLab's MLFlow server.

    Parameters
    ----------
    project_path
        Path to the project where to upload model entries.
    experiment_folder
        Directory in which to upload results from.
    experiment_name
        A string indicating the experiment name (e.g. "exp-pasa-mc" or "exp-densenet-mc-ch").
    run_name
        A string indicating the run name (e.g. "run-1").
    metrics
        List of metrics to upload.
    upload_limit_mb
        Maximim upload size in MB (set to 0 for no limit).
    """

    tmpdir = _assemble_artifacts(experiment_folder, upload_limit_mb)

    experiment_name, run_name = _user_or_default_names(
        experiment_folder, experiment_name, run_name
    )

    _upload_ml_experiment(
        project_path=project_path,
        experiment_name=experiment_name,
        run_name=run_name,
        artifact_path=pathlib.Path(tmpdir.name),
        parameters=_assemble_parameters(experiment_folder),
        metrics=_assemble_metrics(experiment_folder, metrics),
    )
